#!/usr/bin/csi
; vim:ts=2:sw=2:et:
; mysql-tests.scm,v 1.3 2005/08/04 09:14:44 tbutzon Exp

(use mysql)

;;;

(define (collect-until proc pred)
  (do ([val (proc) (proc)]
       [lst '() (cons val lst)])
      [(pred val) lst]) )

(define (collect-while proc pred)
  (collect-until proc (complement pred)) )

;;;

(define (mysql-fetch-field-item conn getter)
  (collect-while (lambda () (mysql-fetch-field-slot conn getter)) identity) )

#; ;UNUSED
(define (mysql-fetch-field-items conn . getters)
  (collect-while (lambda () (apply mysql-fetch-field-slots conn getters)) identity) )

;;;

(print "Opening the connection")
(define *conn* (mysql-connect user: *user* passwd: *passwd* host: *host*))
(print "Database: " *conn*)

(unless *conn*
  (print "Unable to connect to database.")
  (print "You might want to edit \"mysql-tests.scm\" to set *user*, *passwd*, and *host*.")
  (exit) )

(print "MySQL errno: " (mysql-errno *conn*))
(print "MySQL error: " (mysql-error *conn*))
(newline)

(define (mysql-print-and-query *conn* query)
  (print "QUERY: " query)
  (mysql-query *conn* query))

(print "Performing query: SHOW DATABASES")
(mysql-print-and-query *conn* "SHOW DATABASES")
(print "Number of rows: " (mysql-num-rows *conn*))

(newline)
(print "The rows can be enumerated one-per-line, like this:")
(mysql-row-for-each
  *conn*
  (lambda (row row-idx)
    (display (conc row-idx ": " (row "Database") "\n"))))

(newline)
(display "Or they can just be comma-separated, like so:")
(mysql-row-for-each
  *conn*
  (lambda (row row-idx)
    (display (conc (row "Database")
                   (if (< row-idx (mysql-num-rows *conn*)) ", " "\n")))))

(mysql-print-and-query *conn* "DROP DATABASE IF EXISTS mysql_egg_test")

(mysql-print-and-query *conn* "CREATE DATABASE mysql_egg_test")

(print "selecting database: mysql_egg_test")
(mysql-select-db *conn* "mysql_egg_test")

(newline)
(mysql-print-and-query *conn* "CREATE TABLE test1 (
  id INTEGER UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
  testboolnn BOOLEAN NOT NULL,
  testbool BOOLEAN DEFAULT NULL
  )")

(newline)
(mysql-print-and-query *conn*
  "INSERT INTO test1 SET testboolnn=TRUE, testbool=NULL")
(mysql-print-and-query *conn*
  "INSERT INTO test1 SET testboolnn=FALSE, testbool=FALSE")

(newline)
(mysql-print-and-query *conn* "SELECT testboolnn, testbool FROM test1")
(mysql-row-for-each
  *conn*
  (lambda (row idx)
    (print "Row " idx " lengths: " (mysql-fetch-lengths *conn*))))

(newline)
(mysql-print-and-query *conn* "SELECT testboolnn, testbool FROM test1")
(mysql-row-for-each *conn* (lambda (row idx)
  (display (conc "row " idx ": testboolnn=" (row "testboolnn")
                 ", testbool="
                 (row "testbool") "\n"))))

(newline)
(mysql-print-and-query *conn* "SELECT testboolnn, testbool FROM test1")
(let ([names (mysql-fetch-field-item *conn* mysql-field-name)])
  (pretty-print
   (mysql-row-map
     *conn*
     (lambda (row idx)
       (map (lambda (f) (cons f (row f))) names)))) )

;; Finished
(newline)
(print "Closing the connection")
(mysql-close *conn*)
