; vim:ts=2:sw=2:et:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; MySQL Egg for the Chicken Scheme system.
;;
;; mysql.scm,v 1.4 2005/08/04 10:09:26 tbutzon Exp
;;
;; Author: Toby Butzon (toby@butzon.com)
;; Revisions:
;;	 2005.08.04 Initial release.
;;	 2007.01.20 Added -lz for Mac OS X compilation.
;;								(Thanks to Daniel Sadilek.)
;;							Changed #include <mysql/mysql.h> to #include <mysql.h>
;;								(Thanks to Kon Lovett and Daniel Sadilek.)
;;							Removed some misc. compilation warnings.
;;								(Thanks to Kon Lovett.)
;;							Added (error ...) calls instead of silently failing in
;;								mysql-connect.
;;								(Thanks to Daniel Sadilek.)
;;	 2008.01.24 Use of SRFI-12 conditions instead of 'error'.
;;							MYSQL_FIELD support.
;;							MY_CHARSET_INFO support.
;;							More row mapping functions.
;;							Options support
;;							SSL support
;;							Removed deprecated procedures:
;;								mysql-create-db
;;								mysql-drop-db
;;								mysql-eof
;;								mysql-reload
;;							(Kon Lovett)
;;       2009.06.18 Added new procedure mysql-is-closed?
;;                  Connection check in most mysql procedures
;;                  (Daishi Kato)
;;       2009.07.02 Fixed a bug in binary-char-pointer->string
;;                  Fixed a bug in mysql-fetch-lengths(-internal)
;;                  (Daishi Kato)
;;
;; This egg provides the MySQL C API via Chicken's foreign function
;; interface. The function names in the MySQL C API are uniformly
;; exported with "foreign-" prefixed to the name, and underscores are
;; mapped to the more Scheme-ish dash; e.g., the MySQL C API's
;; mysql_query is provided by this egg as foreign-mysql-query. Although
;; these functions are exported, they are not verbosely documented in
;; the egg documentation simply because the MySQL C API is already
;; documented in full on MySQL's website:
;;
;;		http://dev.mysql.com/doc/mysql/en/c.html
;;
;; Further, this egg provides a layer on top of the foreign API to
;; provide a simpler API that is appropriate to Scheme. These functions
;; are _not_ prefixed with "foreign-" -- instead they are named according
;; to the closest MySQL C API analog, if one exists. For example, the
;; provided mysql-connect is briefer, yet provides the same functionality
;; as, the C API sequence mysql_init, mysql_real_connect, and returns a
;; Scheme-friendly record instead of a FFI C pointer.
;;
;; Please report bugs to <toby@butzon.com>.
;;
;; Copyright (c) 2005 Toby Butzon.
;;
;; Permission is hereby granted, free of charge, to any person obtaining a
;; copy of this software and associated documentation files (the "Software"),
;; to deal in the Software without restriction, including without limitation
;; the rights to use, copy, modify, merge, publish, distribute, sublicense,
;; and/or sell copies of the Software, and to permit persons to whom the
;; Software is furnished to do so, subject to the following conditions:
;;
;; The above copyright notice and this permission notice shall be included
;; in all copies or substantial portions of the Software.
;;
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
;; OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
;; ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
;; OTHER DEALINGS IN THE SOFTWARE.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#>#include <mysql.h><#

(require-extension srfi-1 srfi-4 srfi-12)
(require-extension lolevel)

(require-for-syntax 'srfi-13)

(declare
	(usual-integrations)
	(generic)
	(no-procedure-checks-for-usual-bindings)
	(no-bound-checks)
	(inline)
	(unused
		my-charset-info-name-set!
		my-charset-info-csname-set!
		my-charset-info-comment-set!
		my-charset-info-dir-set!
		my-charset-info-mbminlen-set!
		my-charset-info-mbmaxlen-set!
		mysql-field-db-set!
		mysql-field-type-set!
		mysql-field-charsetnr-set!
		mysql-field-catalog-set!
		mysql-field-org-name-set!
		mysql-field-name-set!
		mysql-field-def-set!
		mysql-field-org-table-set!
		mysql-field-table-set!
		mysql-field-flags-set!
		mysql-field-org-table
		mysql-field-decimals-set!
		mysql-field-def-length-set!
		mysql-field-catalog-length-set!
		mysql-field-db-length-set!
		mysql-field-org-table-length-set!
		mysql-field-table-length-set!
		mysql-field-org-name-length-set!
		mysql-field-name-length-set!
		mysql-field-max-length-set!
		mysql-field-length-set! )
	(bound-to-procedure
		mysql-error
		mysql-errno
		#;
		mysql-stmt-close )
	(export
		;; direct api
		foreign-mysql-affected-rows
		foreign-mysql-change-user
		foreign-mysql-character-set-name
		foreign-mysql-close
		foreign-mysql-data-seek
		foreign-mysql-debug
		foreign-mysql-dump-debug-info
		foreign-mysql-errno
		foreign-mysql-error
		foreign-mysql-escape-string
		foreign-mysql-fetch-field
		foreign-mysql-fetch-fields
		foreign-mysql-fetch-field-direct
		foreign-mysql-fetch-lengths
		foreign-mysql-fetch-row
		foreign-mysql-field-count
		foreign-mysql-field-seek
		foreign-mysql-field-tell
		foreign-mysql-free-result
		foreign-mysql-get-character-set-info
		foreign-mysql-get-client-info
		foreign-mysql-get-client-version
		foreign-mysql-get-host-info
		foreign-mysql-get-proto-info
		foreign-mysql-get-server-info
		foreign-mysql-get-server-version
		foreign-mysql-hex-string
		foreign-mysql-info
		foreign-mysql-init
		foreign-mysql-insert-id
		foreign-mysql-kill
		foreign-mysql-library-init
		foreign-mysql-library-end
		foreign-mysql-list-dbs
		foreign-mysql-list-fields
		foreign-mysql-list-processes
		foreign-mysql-list-tables
		foreign-mysql-num-fields
		foreign-mysql-num-rows
		foreign-mysql-options
		foreign-mysql-ping
		foreign-mysql-query
		foreign-mysql-real-connect
		foreign-mysql-real-escape-string
		foreign-mysql-real-query
		foreign-mysql-row-seek
		foreign-mysql-row-tell
		foreign-mysql-select-db
		foreign-mysql-set-character-set
		foreign-mysql-set-server-option
		foreign-mysql-shutdown
		foreign-mysql-sqlstate
		foreign-mysql-ssl-set
		foreign-mysql-stat
		foreign-mysql-store-result
		foreign-mysql-thread-id
		foreign-mysql-use-result
		foreign-mysql-warning-count
		foreign-mysql-commit
		foreign-mysql-rollback
		foreign-mysql-autocommit
		foreign-mysql-more-results
		foreign-mysql-next-result

		;; scheme api
		mysql-connection?
		mysql-affected-rows
		mysql-change-user
		mysql-character-set-name
		mysql-close ; customized
		mysql-connect ; customized
		#;mysql-data-seek ; omitted (low level)
		mysql-debug
		mysql-dump-debug-info
		mysql-errno
		mysql-error ; customized
		mysql-escape-string ; customized
		mysql-fetch-field
		mysql-fetch-fields
		mysql-fetch-field-direct
		mysql-fetch-lengths ; customized
		mysql-fetch-row ; customized
		mysql-field-count
		#;mysql-field-seek ; omitted (low level)
		#;mysql-field-tell ; omitted (low level)
		mysql-free-result
		mysql-get-character-set-info
		mysql-get-client-info
		mysql-get-client-version
		mysql-get-host-info
		mysql-get-proto-info
		mysql-get-server-info
		mysql-get-server-version
		#;mysql-hex-string ; omitted (too new)
		mysql-info
		#;mysql-init ; omitted (low level)
		mysql-insert-id
		mysql-kill
		#;mysql-library-init ; omitted (too new)
		#;mysql-library-end ; omitted (too new)
		mysql-list-dbs
		mysql-list-fields ; (nearly deprecated)
		mysql-list-processes
		mysql-list-tables
		mysql-num-fields
		mysql-num-rows
		mysql-ping
		mysql-query
		#;mysql-real-connect ; omitted (use mysql-connect)
		#;mysql-real-escape-string ; omitted (use mysql-escape-string)
		#;mysql-real-query ; omitted (use mysql-query)
		#;mysql-row-seek ; omitted (low level)
		#;mysql-row-tell ; omitted (low level)
		mysql-select-db
		mysql-set-character-set
		#;mysql-set-server-option ; omitted (too new)
		#;mysql-shutdown ; omitted (too new)
		#;mysql-sqlstate ; omitted (too new)
		mysql-stat
		mysql-store-result
		mysql-thread-id
		#;mysql-use-result ; omitted (doesn't fit with Scheme layer; see mysql-query)
		#;mysql-warning-count ; omitted (too new)
		#;mysql-commit ; omitted (too new)
		#;mysql-rollback ; omitted (too new)
		#;mysql-autocommit ; omitted (too new)
		#;mysql-more-results ; omitted (too new)
		#;mysql-next-result ; omitted (too new)

		;;
		make-mysql-options

		;; ssl parameters api
		make-mysql-ssl
		mysql-ssl?
		mysql-ssl-key-pathname
		mysql-ssl-certificate-pathname
		mysql-ssl-certificate-authority-pathname
		mysql-ssl-trusted-certificates-pathname
		mysql-ssl-ciphers

		;; "extended" api
		mysql-rewind

		;; "map" api
		mysql-row-fold
		mysql-row-for-each
		mysql-row-map
		mysql-query-fold
		mysql-query-for-each
		mysql-query-map
		; synonyms
		mysql-foreach-row
		mysql-query-foreach

		;; MY_CHARSET_INFO api
		; slot getters
		my-charset-info-name
		my-charset-info-csname
		my-charset-info-comment
		my-charset-info-dir
		my-charset-info-mbminlen
		my-charset-info-mbmaxlen

		;; MYSQL_FIELD api
		; predicates
		mysql-field-flags-on?
		mysql-field-flags-off?
		mysql-field-primary-key?
		mysql-field-not-null?
		mysql-field-numeric?
		mysql-field-type-any?
		mysql-field-type=?
		mysql-field-type-clock?
		mysql-field-type-number?
		mysql-field-type-string?
		mysql-field-type-blob?
		mysql-field-type-magnitude?
		mysql-field-type-binary?
		mysql-field-type-text?
		;
		mysql-fetch-field-specific
		; multi-field getters
		mysql-field-slots
		mysql-fetch-field-slots-direct
		mysql-fetch-field-slots
		mysql-fetch-field-slot-direct
		mysql-fetch-field-slot
		; slot getters
		mysql-field-name
		mysql-field-org-name
		mysql-field-table
		mysql-field-org-table
		mysql-field-db
		mysql-field-catalog
		mysql-field-def
		mysql-field-length
		mysql-field-max-length
		mysql-field-name-length
		mysql-field-org-name-length
		mysql-field-table-length
		mysql-field-org-table-length
		mysql-field-db-length
		mysql-field-catalog-length
		mysql-field-def-length
		mysql-field-flags
		mysql-field-decimals
		mysql-field-charsetnr
		mysql-field-type
		
		; check if connection is closed
		mysql-is-closed?

		;; connection client flags
		; values
		client-long-password
		client-long-flag
		client-connect-with-db
		client-protocol-41
		client-secure-connection
		client-transactions
		client-compress
		client-found-rows
		client-ignore-sigpipe
		client-ignore-space
		client-interactive
		client-local-files
		client-multi-results
		client-multi-statements
		client-no-schema
		client-odbc
		client-ssl
		; converters
		mysql-client-flags-value
		mysql-client-flags-symbol

		;; enum enum_mysql_set_option
		; values
		mysql-option-multi-statements-on
		mysql-option-multi-statements-off
		; converters
		mysql-server-option-value
		mysql-server-option-symbol

		;; enum mysql_option
		; values
		mysql-opt-connect-timeout
		mysql-opt-compress
		mysql-opt-named-pipe
		mysql-init-command
		mysql-read-default-file
		mysql-read-default-group
		mysql-set-charset-dir
		mysql-set-charset-name
		mysql-opt-local-infile
		mysql-opt-protocol
		mysql-shared-memory-base-name
		mysql-opt-read-timeout
		mysql-opt-write-timeout
		mysql-opt-use-result
		mysql-opt-use-remote-connection
		mysql-opt-use-embedded-connection
		mysql-opt-guess-connection
		mysql-set-client-ip
		mysql-secure-auth
		mysql-report-data-truncation
		; converters
		mysql-option-value
		mysql-option-symbol

		;; enum enum_field_types
		; values
		mysql-type-decimal
		mysql-type-tiny
		mysql-type-short
		mysql-type-long
		mysql-type-int24
		mysql-type-float
		mysql-type-double
		mysql-type-null
		mysql-type-timestamp
		mysql-type-longlong
		mysql-type-date
		mysql-type-time
		mysql-type-datetime
		mysql-type-year
		mysql-type-newdate
		mysql-type-varchar
		mysql-type-bit
		mysql-type-newdecimal
		mysql-type-enum
		mysql-type-set
		mysql-type-tiny-blob
		mysql-type-medium-blob
		mysql-type-long-blob
		mysql-type-blob
		mysql-type-var-string
		mysql-type-string
		mysql-type-geometry
		; converters
		mysql-type-value
		mysql-type-symbol

		;; field flags
		; values
		not-null-flag
		pri-key-flag
		unique-key-flag
		multiple-key-flag
		unsigned-flag
		zerofill-flag
		binary-flag
		auto-increment-flag
		no-default-value-flag
		; deprecated
		enum-flag
		set-flag
		blob-flag
		timestamp-flag
		; converters
		mysql-field-flags-value
		mysql-field-flags-symbol ) )

;;;

(define-for-syntax (c-name->scheme-name str)
	(string-downcase (string-translate str "_" "-")) )

;;;

;=======================================================================
; Interface to C API
;
; The entire C API is mapped using Chicken's foreign function interface.
;

;-----------------------------------------------------------------------
; Foreign type definitions.
;

(define-foreign-type mysql-ptr (c-pointer "MYSQL"))
(define-foreign-type mysql-res-ptr (c-pointer "MYSQL_RES"))
(define-foreign-type mysql-rows-ptr (c-pointer "MYSQL_ROWS"))
(define-foreign-type mysql-field-ptr (c-pointer "MYSQL_FIELD"))

;; MYSQL_ROW (we just pass it around, not def-ref'ed)

(define-foreign-type mysql-row (c-pointer (c-pointer char)))

;; my_ulonglong

(define-foreign-type my-ulonglong number)

;; my_bool

;(define-foreign-type my-bool "my_bool" (lambda (x) (if x 1 0)) (lambda (x) (not (fx= 0 x))))
;the above raises an error in csc, and the below is a workaround.
(define-foreign-type my-bool char (lambda (x) (if x 1 0)) (lambda (x) (not (fx= 0 x))))

;; MYSQL_FIELD_OFFSET

(define-foreign-type mysql-field-offset unsigned-int)

;-----------------------------------------------------------------------
; Enumeration & flag value constants
;

;;

(define-macro (gen-public-enum ?typ . ?syms)
	` ; Begin Result
	(begin
		,@(map (lambda (sym)
             (if (pair? sym)
                 `(define ,(car sym) ,(cadr sym))
                 `(define ,(string->symbol (c-name->scheme-name (symbol->string sym))) ,sym)))
			     ?syms)
    (define (,(string->symbol (string-append (symbol->string ?typ) "-value")) . syms)
      (,(string->symbol (string-append (symbol->string ?typ) "->number")) syms))
    (define (,(string->symbol (string-append (symbol->string ?typ) "-symbol")) val)
      (,(string->symbol (string-append "number->" (symbol->string ?typ))) val))) )

(define-macro ($define-foreign-enum ?tspec . ?rest)
  (let ([1st-item (lambda (x) (if (pair? x) (car x) x))])
    (let-values ([(?flags ?especs)
                  (if (null? ?rest)
                      (syntax-error '$define-foreign-enum "missing enumeration specification")
                      (if (boolean? (car ?rest))
                          (values (list (car ?rest)) (cdr ?rest))
                          (values '() ?rest) ) ) ] )
      (let ([?typ (1st-item ?tspec)]
            [?enams (map 1st-item ?especs) ] )
        ` ; Begin Result
        (begin
          (define-foreign-enum ,?tspec ,@?flags ,@?especs)
          (gen-public-enum ,?typ ,@?enams) ) ) ) ) )

;;

($define-foreign-enum (mysql-server-option (enum "enum_mysql_set_option"))
	#f	; No aliases!
	MYSQL_OPTION_MULTI_STATEMENTS_ON
	MYSQL_OPTION_MULTI_STATEMENTS_OFF )

($define-foreign-enum (mysql-option (enum "mysql_option"))
	#f	; No aliases!
	MYSQL_OPT_CONNECT_TIMEOUT
	MYSQL_OPT_COMPRESS
	MYSQL_OPT_NAMED_PIPE
	MYSQL_INIT_COMMAND
	MYSQL_READ_DEFAULT_FILE
	MYSQL_READ_DEFAULT_GROUP
	MYSQL_SET_CHARSET_DIR
	MYSQL_SET_CHARSET_NAME
	MYSQL_OPT_LOCAL_INFILE
	MYSQL_OPT_PROTOCOL
	MYSQL_SHARED_MEMORY_BASE_NAME
	MYSQL_OPT_READ_TIMEOUT
	MYSQL_OPT_WRITE_TIMEOUT
	MYSQL_OPT_USE_RESULT
	MYSQL_OPT_USE_REMOTE_CONNECTION
	MYSQL_OPT_USE_EMBEDDED_CONNECTION
	MYSQL_OPT_GUESS_CONNECTION
	MYSQL_SET_CLIENT_IP
	MYSQL_SECURE_AUTH
	MYSQL_REPORT_DATA_TRUNCATION )

($define-foreign-enum (mysql-type (enum "enum_field_types"))
	#f	; No aliases!
	MYSQL_TYPE_DECIMAL
	MYSQL_TYPE_TINY
	MYSQL_TYPE_SHORT
	MYSQL_TYPE_LONG
	MYSQL_TYPE_INT24
	MYSQL_TYPE_FLOAT
	MYSQL_TYPE_DOUBLE
	MYSQL_TYPE_NULL
	MYSQL_TYPE_TIMESTAMP
	MYSQL_TYPE_LONGLONG
	MYSQL_TYPE_DATE
	MYSQL_TYPE_TIME
	MYSQL_TYPE_DATETIME
	MYSQL_TYPE_YEAR
	MYSQL_TYPE_NEWDATE
	MYSQL_TYPE_VARCHAR
	MYSQL_TYPE_BIT
	MYSQL_TYPE_NEWDECIMAL
	MYSQL_TYPE_ENUM
	MYSQL_TYPE_SET
	MYSQL_TYPE_TINY_BLOB
	MYSQL_TYPE_MEDIUM_BLOB
	MYSQL_TYPE_LONG_BLOB
	MYSQL_TYPE_BLOB
	MYSQL_TYPE_VAR_STRING
	MYSQL_TYPE_STRING
	MYSQL_TYPE_GEOMETRY )

($define-foreign-enum (mysql-field-flags unsigned-int)
	#f	; No aliases!
	NOT_NULL_FLAG									; field can't be NULL
	PRI_KEY_FLAG									; field is part of a primary key
	UNIQUE_KEY_FLAG								; field is part of a unique key
	MULTIPLE_KEY_FLAG							; field is part of a key
	UNSIGNED_FLAG									; field is unsigned
	ZEROFILL_FLAG									; field is zerofill
	BINARY_FLAG										; field is binary
	AUTO_INCREMENT_FLAG						; field is a autoincrement field
	NO_DEFAULT_VALUE_FLAG					; field doesn't have default value
	NUM_FLAG											; field is num (for clients)
	; deprecated
	ENUM_FLAG											; field is an enum
	SET_FLAG											; field is a set
	BLOB_FLAG											; field is a blob
	TIMESTAMP_FLAG	)							; field is a timestamp

($define-foreign-enum (mysql-client-flags unsigned-int)
	#f	; No aliases!
	CLIENT_LONG_PASSWORD					; new more secure passwords
	CLIENT_LONG_FLAG							; Get all column flags
	CLIENT_CONNECT_WITH_DB				; One can specify db on connect
	CLIENT_PROTOCOL_41						; New 4.1 protocol
	CLIENT_SECURE_CONNECTION			; New 4.1 authentication
	CLIENT_TRANSACTIONS						; Client knows about transactions
	CLIENT_COMPRESS								; Can use compression protocol
	; Return the number of found (matched) rows, not the number of affected rows.
	CLIENT_FOUND_ROWS
	; Prevents the client library from installing a SIGPIPE signal handler. This
	; can be used to avoid conflicts with a handler that the application has
	; already installed.
	CLIENT_IGNORE_SIGPIPE
	; Allow spaces after function names. Makes all functions names reserved
	; words.
	CLIENT_IGNORE_SPACE
	; Allow interactive_timeout seconds (instead of wait_timeout seconds) of
	; inactivity before closing the connection. The client's session wait_timeout
	; variable is set to the value of the session interactive_timeout variable.
	CLIENT_INTERACTIVE
	; Enable LOAD DATA LOCAL handling.
	CLIENT_LOCAL_FILES
	; Tell the server that the client can handle multiple result sets from
	; multiple-statement executions or stored procedures. This is automatically set
	; if CLIENT_MULTI_STATEMENTS is set. See the note following this table for more
	; information about this flag.
	CLIENT_MULTI_RESULTS
	; Tell the server that the client may send multiple statements in a single
	; string (separated by ;). If this flag is not set, multiple-statement
	; execution is disabled. See the note following this table for more information
	; about this flag.
	CLIENT_MULTI_STATEMENTS
	; Don't allow the db_name.tbl_name.col_name syntax. This is for ODBC. It
	; causes the parser to generate an error if you use that syntax, which is
	; useful for trapping bugs in some ODBC programs.
	CLIENT_NO_SCHEMA
	; Unused.
	CLIENT_ODBC
	; Use SSL (encrypted protocol). This option should not be set by application
	; programs; it is set internally in the client library. Instead, use
	; mysql_ssl_set() before calling mysql_real_connect().
	CLIENT_SSL )

;-----------------------------------------------------------------------
; C Structures
;

;; not "mysql-field-ptr", named so getters have nice name.

(define-foreign-record (mysql-field "MYSQL_FIELD")
	(rename: c-name->scheme-name)
	; No ctor or dtor
	(c-string name)												; Name of column
	(c-string org_name)										; Original column name, if an alias
	(c-string table)											; Table of column if column was a field
	(c-string org_table)									; Org table name, if table was an alias
	(c-string db)													; Database for table
	(c-string catalog)										; Catalog for table
	(c-string def)												; Default value (set by mysql_list_fields)
	(unsigned-long length)								; Width of column (create length)
	(unsigned-long max_length)						; Max width for selected set
	(unsigned-integer name_length)
	(unsigned-integer org_name_length)
	(unsigned-integer table_length)
	(unsigned-integer org_table_length)
	(unsigned-integer db_length)
	(unsigned-integer catalog_length)
	(unsigned-integer def_length)
	(unsigned-integer flags)							; Div flags
	(unsigned-integer decimals)						; Number of decimals in field
	(unsigned-integer charsetnr)					; Character set
	(mysql-type type) )										; Type of field. See mysql_com.h for types

;; not "my-charset-info-ptr", named so getters have nice name.

(define-foreign-record (my-charset-info "MY_CHARSET_INFO")
	(rename: c-name->scheme-name)
	(constructor: allocate-my-charset-info)
	(destructor: free-my-charset-info)
	(c-string name)								; character set name
	(c-string csname)							; collation name
	(c-string comment)						; comment
	(c-string dir)								; directory
	(unsigned-integer mbminlen)		; multi byte character min. length
	(unsigned-integer mbmaxlen) ) ; multi byte character max. length

;-----------------------------------------------------------------------
; Foreign function definitions for the MySQL C API using the
; MySQL "my_ulonglong" type.
;

; 24.2.3.1. mysql_affected_rows()
; my_ulonglong mysql_affected_rows(MYSQL *mysql)
(define foreign-mysql-affected-rows
	(foreign-lambda* my-ulonglong ((mysql-ptr mysql))
	 "return( mysql_affected_rows( mysql ) );"))

; 24.2.3.34. mysql_insert_id()
; my_ulonglong mysql_insert_id(MYSQL *mysql)
(define foreign-mysql-insert-id
	(foreign-lambda* my-ulonglong ((mysql-ptr mysql))
	 "return( mysql_insert_id( mysql ) );"))

; 24.2.3.43. mysql_num_rows()
; my_ulonglong mysql_num_rows(MYSQL_RES *result)
(define foreign-mysql-num-rows
	(foreign-lambda* my-ulonglong ((mysql-res-ptr result))
	 "return( mysql_num_rows( result ) );"))

; 24.2.3.7. mysql_data_seek()
; void mysql_data_seek(MYSQL_RES *result, my_ulonglong offset)
(define foreign-mysql-data-seek
	(foreign-lambda* void ((mysql-res-ptr result) (my-ulonglong offset))
	 "mysql_data_seek( result, offset );"))

;-----------------------------------------------------------------------
; Foreign function definitions for mysql_options from MySQL C API.
;

;;

(define foreign-mysqlaux-options-none
	(foreign-lambda* int ((mysql-ptr mysql) (mysql-option opt))
   "return( mysql_options( mysql, opt, NULL ) );"))

;;

(define foreign-mysqlaux-options-string
	(foreign-lambda* int ((mysql-ptr mysql) (mysql-option opt) (c-string value))
   "return( mysql_options( mysql, opt, value ) );"))

;;

(define foreign-mysqlaux-options-ulong
	(foreign-lambda* int ((mysql-ptr mysql) (mysql-option opt) (unsigned-long value))
   "return( mysql_options( mysql, opt, &value ) );"))

;-----------------------------------------------------------------------
; Foreign function definitions for mysqlaux functions.
;

;;

(define foreign-mysqlaux-field-index
	(foreign-lambda* int ((mysql-res-ptr result) (nonnull-c-string name)
	                      (unsigned-integer num_fields))
#<<EOS
	MYSQL_FIELD *fields = mysql_fetch_fields( result );
	unsigned int i;

	for (i = 0; i < num_fields; i++) {
		if (0 == strcasecmp( name, fields[ i ].name )) {
			return( i );
		}
	}

	return( -1 );
EOS
  ) )

;;

(define foreign-mysqlaux-is-binary-field
	(foreign-lambda* bool ((mysql-res-ptr result) (unsigned-integer fldidx))
#<<EOS
	MYSQL_FIELD *fields = mysql_fetch_fields( result );

	switch (fields[fldidx].type) {
		case MYSQL_TYPE_VARCHAR:
		case MYSQL_TYPE_TINY_BLOB:
		case MYSQL_TYPE_MEDIUM_BLOB:
		case MYSQL_TYPE_LONG_BLOB:
		case MYSQL_TYPE_BLOB:
		case MYSQL_TYPE_VAR_STRING:
		case MYSQL_TYPE_STRING:
			return( 63 == fields[ fldidx ].charsetnr );
		default:
		  break;
	}

	return( 0 );
EOS
  ) )

;;

(define foreign-mysqlaux-fetch-column-string-direct
	(foreign-lambda* c-string ((mysql-row row) (unsigned-integer fldidx))
	 "return( row[ fldidx ] );" ) )

;;

(define foreign-mysqlaux-fetch-column-data-direct
	(foreign-lambda* c-pointer ((mysql-row row) (unsigned-integer fldidx))
	 "return( row[ fldidx ] );" ) )

;-----------------------------------------------------------------------
; Foreign function definitions from MySQL C API.
;
; I've copied the listing of MySQL C API functions straight from the
; MySQL manual. They are in alphabetical order, exactly as they appear
; in the manual. Further, the C function signature is copied below the
; manual entry. Finally, the Scheme foreign lambda mapping follows the
; C function signature.
;

; 24.2.3.2. mysql_change_user()
; my_bool mysql_change_user(MYSQL *mysql, const char *user,
;		const char *password, const char *db)
(define foreign-mysql-change-user
	(foreign-lambda my-bool "mysql_change_user" mysql-ptr c-string c-string
									c-string))

; 24.2.3.3. mysql_character_set_name()
; const char *mysql_character_set_name(MYSQL *mysql)
(define foreign-mysql-character-set-name
	(foreign-lambda c-string "mysql_character_set_name" mysql-ptr))

; 24.2.3.4. mysql_close()
; void mysql_close(MYSQL *mysql)
(define foreign-mysql-close
	(foreign-lambda void "mysql_close" mysql-ptr))

; 24.2.3.8. mysql_debug()
; void mysql_debug(const char *debug)
(define foreign-mysql-debug
	(foreign-lambda void "mysql_debug" c-string))

; 24.2.3.10. mysql_dump_debug_info()
; int mysql_dump_debug_info(MYSQL *mysql)
(define foreign-mysql-dump-debug-info
	(foreign-lambda integer "mysql_dump_debug_info" mysql-ptr))

; 24.2.3.12. mysql_errno()
; unsigned int mysql_errno(MYSQL *mysql)
(define foreign-mysql-errno
	(foreign-lambda unsigned-integer "mysql_errno" mysql-ptr))

; 24.2.3.13. mysql_error()
; const char *mysql_error(MYSQL *mysql)
(define foreign-mysql-error
	(foreign-lambda c-string "mysql_error" mysql-ptr))

; 24.2.3.14. mysql_escape_string()
; unsigned long mysql_escape_string(char *to, const char *from,
;		unsigned long length)
(define foreign-mysql-escape-string
	(foreign-lambda unsigned-long "mysql_escape_string" c-string c-string
									unsigned-long))

; 24.2.3.15. mysql_fetch_field()
; MYSQL_FIELD *mysql_fetch_field(MYSQL_RES *result)
(define foreign-mysql-fetch-field
	(foreign-lambda mysql-field-ptr "mysql_fetch_field" mysql-res-ptr))

; 24.2.3.16. mysql_fetch_fields()
; MYSQL_FIELD *mysql_fetch_fields(MYSQL_RES *result)
(define foreign-mysql-fetch-fields
	(foreign-lambda mysql-field-ptr "mysql_fetch_fields" mysql-res-ptr))

; 24.2.3.17. mysql_fetch_field_direct()
; MYSQL_FIELD *mysql_fetch_field_direct(MYSQL_RES *result, unsigned int fieldnr)
(define foreign-mysql-fetch-field-direct
	(foreign-lambda mysql-field-ptr "mysql_fetch_field_direct" mysql-res-ptr
									unsigned-integer))

; 24.2.3.18. mysql_fetch_lengths()
; unsigned long *mysql_fetch_lengths(MYSQL_RES *result)
(define foreign-mysql-fetch-lengths
	(foreign-lambda (c-pointer unsigned-long) "mysql_fetch_lengths" mysql-res-ptr))

; 24.2.3.19. mysql_fetch_row()
; MYSQL_ROW mysql_fetch_row(MYSQL_RES *result)
(define foreign-mysql-fetch-row
	(foreign-lambda mysql-row "mysql_fetch_row" mysql-res-ptr))

; 24.2.3.20. mysql_field_count()
; unsigned int mysql_field_count(MYSQL *mysql)
(define foreign-mysql-field-count
	(foreign-lambda unsigned-integer "mysql_field_count" mysql-ptr))

; 24.2.3.21. mysql_field_seek()
; MYSQL_FIELD_OFFSET mysql_field_seek(MYSQL_RES *result,
;		MYSQL_FIELD_OFFSET offset)
(define foreign-mysql-field-seek
	(foreign-lambda mysql-field-offset "mysql_field_seek" mysql-res-ptr
									mysql-field-offset))

; 24.2.3.22. mysql_field_tell()
; MYSQL_FIELD_OFFSET mysql_field_tell(MYSQL_RES *result)
(define foreign-mysql-field-tell
	(foreign-lambda mysql-field-offset "mysql_field_tell" mysql-res-ptr))

; 24.2.3.23. mysql_free_result()
; void mysql_free_result(MYSQL_RES *result)
(define foreign-mysql-free-result
	(foreign-lambda void "mysql_free_result" mysql-res-ptr))

; 24.2.3.26.mysql_get_character_set_info()
; void mysql_get_character_set_info(MYSQL *mysql, MY_CHARSET_INFO *cs)
(define foreign-mysql-get-character-set-info
	(foreign-lambda void "mysql_get_character_set_info" mysql-ptr my-charset-info))

; 24.2.3.25. mysql_get_client_info()
; char *mysql_get_client_info(void)
(define foreign-mysql-get-client-info
	(foreign-lambda c-string "mysql_get_client_info"))

; 24.2.3.26. mysql_get_client_version()
; unsigned long mysql_get_client_version(void)
(define foreign-mysql-get-client-version
	(foreign-lambda unsigned-long "mysql_get_client_version"))

; 24.2.3.27. mysql_get_host_info()
; char *mysql_get_host_info(MYSQL *mysql)
(define foreign-mysql-get-host-info
	(foreign-lambda c-string "mysql_get_host_info" mysql-ptr))

; 24.2.3.28. mysql_get_proto_info()
; unsigned int mysql_get_proto_info(MYSQL *mysql)
(define foreign-mysql-get-proto-info
	(foreign-lambda unsigned-integer "mysql_get_proto_info" mysql-ptr))

; 24.2.3.29. mysql_get_server_info()
; char *mysql_get_server_info(MYSQL *mysql)
(define foreign-mysql-get-server-info
	(foreign-lambda c-string "mysql_get_server_info" mysql-ptr))

; 24.2.3.30. mysql_get_server_version()
; unsigned long mysql_get_server_version(MYSQL *mysql)
(define foreign-mysql-get-server-version
	(foreign-lambda unsigned-long "mysql_get_server_version" mysql-ptr))

; 24.2.3.31. mysql_hex_string()
; unsigned long mysql_hex_string(char *to, const char *from,
;		unsigned long length)
(define foreign-mysql-hex-string
	(foreign-lambda unsigned-long "mysql_hex_string" c-string c-string
									unsigned-long))

; 24.2.3.32. mysql_info()
; char *mysql_info(MYSQL *mysql)
(define foreign-mysql-info
	(foreign-lambda c-string "mysql_info" mysql-ptr))

; 24.2.3.33. mysql_init()
; MYSQL *mysql_init(MYSQL *mysql)
(define foreign-mysql-init
	(foreign-lambda mysql-ptr "mysql_init" mysql-ptr))

; 24.2.3.35. mysql_kill()
; int mysql_kill(MYSQL *mysql, unsigned long pid)
(define foreign-mysql-kill
	(foreign-lambda integer "mysql_kill" mysql-ptr unsigned-long))

; 24.2.3.36. mysql_library_init()
; int mysql_library_init(int argc, char **argv, char **groups)
(define foreign-mysql-library-init
	(foreign-lambda integer "mysql_library_init" integer c-pointer c-pointer))

; 24.2.3.37. mysql_library_end()
; void mysql_library_end(void)
(define foreign-mysql-library-end
	(foreign-lambda void "mysql_library_end"))

; 24.2.3.38. mysql_list_dbs()
; MYSQL_RES *mysql_list_dbs(MYSQL *mysql, const char *wild)
(define foreign-mysql-list-dbs
	(foreign-lambda mysql-res-ptr "mysql_list_dbs" mysql-ptr c-string))

; 24.2.3.39. mysql_list_fields()
; MYSQL_RES *mysql_list_fields(MYSQL *mysql, const char *table,
;		const char *wild)
(define foreign-mysql-list-fields
	(foreign-lambda mysql-res-ptr "mysql_list_fields" mysql-ptr c-string
									c-string))

; 24.2.3.40. mysql_list_processes()
; MYSQL_RES *mysql_list_processes(MYSQL *mysql)
(define foreign-mysql-list-processes
	(foreign-lambda mysql-res-ptr "mysql_list_processes" mysql-ptr))

; 24.2.3.41. mysql_list_tables()
; MYSQL_RES *mysql_list_tables(MYSQL *mysql, const char *wild)
(define foreign-mysql-list-tables
	(foreign-lambda mysql-res-ptr "mysql_list_tables" mysql-ptr c-string))

; 24.2.3.42. mysql_num_fields()
; unsigned int mysql_num_fields(MYSQL_RES *result)
(define foreign-mysql-num-fields
	(foreign-lambda unsigned-integer "mysql_num_fields" mysql-res-ptr))

; 24.2.3.44. mysql_options()
; int mysql_options(MYSQL *mysql, enum mysql_option option, const char *arg)
(define foreign-mysql-options
	(foreign-lambda integer "mysql_options" mysql-ptr mysql-option c-pointer))

; 24.2.3.45. mysql_ping()
; int mysql_ping(MYSQL *mysql)
(define foreign-mysql-ping
	(foreign-lambda integer "mysql_ping" mysql-ptr))

; 24.2.3.46. mysql_query()
; int mysql_query(MYSQL *mysql, const char *stmt_str)
;
; NOTE: use "mysql_real_query" instead
(define foreign-mysql-query
	(foreign-lambda integer "mysql_query" mysql-ptr c-string))

; 24.2.3.47. mysql_real_connect()
; MYSQL *mysql_real_connect(MYSQL *mysql, const char *host, const char *user,
;		const char *passwd, const char *db, unsigned int port,
;		const char *unix_socket, unsigned long client_flag)
(define foreign-mysql-real-connect
	(foreign-lambda mysql-ptr "mysql_real_connect"
									mysql-ptr c-string c-string
									c-string c-string unsigned-integer
									c-string unsigned-long))

; 24.2.3.48. mysql_real_escape_string()
; unsigned long mysql_real_escape_string(MYSQL *mysql, char *to,
;		const char *from, unsigned long length)
(define foreign-mysql-real-escape-string
	(foreign-lambda unsigned-long "mysql_real_escape_string" mysql-ptr
									c-string c-string unsigned-long))

; 24.2.3.49. mysql_real_query()
; int mysql_real_query(MYSQL *mysql, const char *query, unsigned long length)
(define foreign-mysql-real-query
	(foreign-lambda unsigned-integer "mysql_real_query" mysql-ptr c-string
									unsigned-long))

; 24.2.3.51. mysql_row_seek()
; MYSQL_ROW_OFFSET mysql_row_seek(MYSQL_RES *result, MYSQL_ROW_OFFSET offset)
(define foreign-mysql-row-seek
	(foreign-lambda mysql-rows-ptr "mysql_row_seek" mysql-res-ptr mysql-rows-ptr))

; 24.2.3.52. mysql_row_tell()
; MYSQL_ROW_OFFSET mysql_row_tell(MYSQL_RES *result)
(define foreign-mysql-row-tell
	(foreign-lambda mysql-rows-ptr "mysql_row_tell" mysql-res-ptr))

; 24.2.3.53. mysql_select_db()
; int mysql_select_db(MYSQL *mysql, const char *db)
(define foreign-mysql-select-db
	(foreign-lambda integer "mysql_select_db" mysql-ptr c-string))

; 24.2.3.54. mysql_set_character_set()
; int mysql_set_character_set(MYSQL *mysql, char *csname)
(define foreign-mysql-set-character-set
	(foreign-lambda integer "mysql_set_character_set" mysql-ptr c-string))

; 24.2.3.55. mysql_set_server_option()
; int mysql_set_server_option(MYSQL *mysql, enum enum_mysql_set_option option)
(define foreign-mysql-set-server-option
	(foreign-lambda integer "mysql_set_server_option" mysql-ptr mysql-server-option))

; 24.2.3.56. mysql_shutdown()
; int mysql_shutdown(MYSQL *mysql, enum enum_shutdown_level shutdown_level)
(define foreign-mysql-shutdown
	(foreign-lambda integer "mysql_shutdown" mysql-ptr integer))

; 24.2.3.57. mysql_sqlstate()
; const char *mysql_sqlstate(MYSQL *mysql)
(define foreign-mysql-sqlstate
	(foreign-lambda c-string "mysql_sqlstate" mysql-ptr))

; 24.2.3.58. mysql_ssl_set()
; int mysql_ssl_set(MYSQL *mysql, const char *key, const char *cert,
;		const char *ca, const char *capath, const char *cipher)
(define foreign-mysql-ssl-set
	(foreign-lambda integer "mysql_ssl_set" mysql-ptr c-string c-string
									c-string c-string c-string))

; 24.2.3.59. mysql_stat()
; char *mysql_stat(MYSQL *mysql)
(define foreign-mysql-stat
	(foreign-lambda c-string "mysql_stat" mysql-ptr))

; 24.2.3.60. mysql_store_result()
; MYSQL_RES *mysql_store_result(MYSQL *mysql)
(define foreign-mysql-store-result
	(foreign-lambda mysql-res-ptr "mysql_store_result" mysql-ptr))

; 24.2.3.61. mysql_thread_id()
; unsigned long mysql_thread_id(MYSQL *mysql)
(define foreign-mysql-thread-id
	(foreign-lambda unsigned-long "mysql_thread_id" mysql-ptr))

; 24.2.3.62. mysql_use_result()
; MYSQL_RES *mysql_use_result(MYSQL *mysql)
(define foreign-mysql-use-result
	(foreign-lambda mysql-res-ptr "mysql_use_result" mysql-ptr))

; 24.2.3.63. mysql_warning_count()
; unsigned int mysql_warning_count(MYSQL *mysql)
(define foreign-mysql-warning-count
	(foreign-lambda unsigned-integer "mysql_warning_count" mysql-ptr))

; 24.2.3.64. mysql_commit()
; my_bool mysql_commit(MYSQL *mysql)
(define foreign-mysql-commit
	(foreign-lambda my-bool "mysql_commit" mysql-ptr))

; 24.2.3.65. mysql_rollback()
; my_bool mysql_rollback(MYSQL *mysql)
(define foreign-mysql-rollback
	(foreign-lambda my-bool "mysql_rollback" mysql-ptr))

; 24.2.3.66. mysql_autocommit()
; my_bool mysql_autocommit(MYSQL *mysql, my_bool mode)
(define foreign-mysql-autocommit
	(foreign-lambda my-bool "mysql_autocommit" mysql-ptr my-bool))

; 24.2.3.67. mysql_more_results()
; my_bool mysql_more_results(MYSQL *mysql)
(define foreign-mysql-more-results
	(foreign-lambda my-bool "mysql_more_results" mysql-ptr))

; 24.2.3.68. mysql_next_result()
; int mysql_next_result(MYSQL *mysql)
(define foreign-mysql-next-result
	(foreign-lambda integer "mysql_next_result" mysql-ptr))

;-----------------------------------------------------------------------
; MySQL helpers
;

;; record printer helper

(define (optional-value->string val lbl #!optional (tst val))
	(if tst
			(conc #\space lbl #\: #\space #\" val #\")
			"") )

;;

(define (unsigned-long-array->u32vector ulptr cnt)
	(let* ([siz (fx* cnt (foreign-value "sizeof( unsigned long )" int))]
				 [store (make-blob siz)])
		(move-memory! ulptr store siz)
		(blob->u32vector/shared store) ) )

;;

#; ; UNUSED
(define char-pointer->string
	(foreign-lambda* c-string ((c-pointer chrptr))
		"return( (char *) chrptr );") )

;;

(define (binary-char-pointer->string chrptr size)
  (if (zero? size)
      ""
      (let ([blob (make-blob size)])
	(move-memory! chrptr blob size)
	(blob->string blob) ) ) )

;; Returns index for field identifier
;; field  : a field index or a field name (converted to a string)
;; resptr : result or #f
;; fldcnt : total number of fields (for check)

(define (mysql-get-field-index resptr field fldcnt)
	(and-let* ([fldidx
	            (cond [(number? field)
                      field]
                    [resptr
                      (foreign-mysqlaux-field-index resptr (->string field) fldcnt)]
                    [else
                      #f])])
		(and (<= 0 fldidx) (< fldidx fldcnt)
				 fldidx ) ) )

;-----------------------------------------------------------------------
; MySQL exceptions
;

;; Returns a Chicken "exn" exception object

(define (make-exn-condition loc msg . args)
	(make-property-condition 'exn 'location loc 'message msg 'arguments args) )

;; Returns a Chicken "mysql" exception object

(define (make-mysql-condition)
	(make-property-condition 'mysql) )

;; Returns a Chicken "exn mysql" exception object

(define (make-exn-mysql-condition loc msg . args)
	(make-composite-condition
	 (apply make-exn-condition loc msg args)
	 (make-mysql-condition)) )

;; Raises a Chicken "exn mysql" exception object

(define (signal-mysql-condition loc msg . args)
	(signal (apply make-exn-mysql-condition loc msg args)) )

;; Raises a Chicken "exn mysql" exception object with a connection error message

(define (signal-mysql-error loc conn . args)
	(let ([msg (or (mysql-error conn)
								 (mysql-errno conn))])
		(apply signal-mysql-condition loc msg args) ) )

;=======================================================================
; Provided Scheme API.
;
; This is an attempt at a Schemer-friendly API to MySQL. Much of the API
; is the same, but the C API has been simplified where possible, and a
; few additional features have been layered on.
;

;-----------------------------------------------------------------------
; MySQL "SSL" record type.
;
; See <http://www.openssl.org/docs/apps/ciphers.html> for a discussion
; of the cipher-list format.
;

;;

(define-record-type mysql-ssl
	(%make-mysql-ssl key cert ca capath cipher)
	mysql-ssl?
	(key mysql-ssl-key-pathname)
	(cert mysql-ssl-certificate-pathname)
	(ca mysql-ssl-certificate-authority-pathname)
	(capath mysql-ssl-trusted-certificates-pathname)
	(cipher mysql-ssl-ciphers) )

;;

(define (make-mysql-ssl #!key key certificate certificate-authority
															trusted-certificates ciphers)
	(%make-mysql-ssl key certificate certificate-authority trusted-certificates ciphers) )

;;

(define-record-printer (mysql-ssl ssl out)
	(let ([key (mysql-ssl-key-pathname ssl)]
				[cert (mysql-ssl-certificate-pathname ssl)]
				[ca (mysql-ssl-certificate-authority-pathname ssl)]
				[capath (mysql-ssl-trusted-certificates-pathname ssl)]
				[cipher (mysql-ssl-ciphers ssl)])
		(display
		 (string-append
			"#<mysql-ssl"
			(optional-value->string key     "key")
			(optional-value->string cert    "cert")
			(optional-value->string ca      "ca")
			(optional-value->string capath  "capath")
			(optional-value->string cipher  "cipher")
			">")
		 out) ) )

;;

(define (mysql-ssl-set! mysql ssl)
  (when ssl
    (foreign-mysql-ssl-set mysql
                           (mysql-ssl-key-pathname ssl)
                           (mysql-ssl-certificate-pathname ssl)
                           (mysql-ssl-certificate-authority-pathname ssl)
                           (mysql-ssl-trusted-certificates-pathname ssl)
                           (mysql-ssl-ciphers ssl)) ) )

;-----------------------------------------------------------------------
; MySQL connection options.
;

;; Returns a mysql-options object

(define (make-mysql-options . opts)
	(let loop ([opts opts] [alst '()])
		(if (null? opts)
				alst
				(let* ([opt (car opts)]
							 [nxt (cdr opts)]
							 [val (if (null? nxt)
												(error 'make-options "missing value for option" opt)
												(car nxt))])
					(unless (number? opt)
						(error 'make-options "invalid option" opt) )
					(unless (or (number? val) (string? val) (not val) (null? val))
						(error 'make-options "invalid option value" val) )
					(loop (cdr nxt) (alist-cons opt val alst)) ) ) ) )

;;

(define (mysql-option-set! mysql opt val)
	(cond [(null? val)
					(foreign-mysqlaux-options-none mysql opt)]
				[(string? val)
					(foreign-mysqlaux-options-string mysql opt val)]
				[(number? val)
					(foreign-mysqlaux-options-ulong mysql opt val)]
				[else
					1 ] ) )

;;

(define (mysql-options-set! mysql options)
  (when options
    (for-each
     (lambda (optitm)
       (let ([opt (car optitm)]
             [val (cdr optitm)])
         (unless (fx= 0 (mysql-option-set! mysql opt val))
           (signal-mysql-condition 'mysql-connect "unknown option" opt val))))
     options) ) )

;-----------------------------------------------------------------------
; MySQL "Connection" record type definition.
;
; I've stuffed the raw FFI pointer into a slot in the mysql-connection
; record. The record is here for a few reasons:
;
;		1) Instead of an ugly #<pointer>, I've defined a pretty printer
;			 to demonstrate that we've actually got a MySQL connection.
;		2) The C API is somewhat more verbose than what normal usage would
;			 need. (For example, usually you don't care whether results are
;			 all read into memory as fast as possible, or if they're read from
;			 the network one-by-one. Thus, the mysql-query function provided
;			 automatically reads the results into memory. For finer granularity,
;			 you're always free to write your own version to use the "raw"
;			 foreign-* functions. I suppose a contribution to determine this
;			 behavior via a (make-parameter ...) parameter may also be
;			 accepted. ;)) Slots are provided in the mysql-connection record
;			 type to allow for this sort of simplifying behavior.
;
; All of the "Scheme API" MySQL functions take instances of this record
; type, instead of a raw FFI pointer (as the foreign-* functions require).
;

;;

(define-record-type mysql-connection
	(%make-mysql-connection host user passwd db port unix-socket
											  	client-flag connptr result result-start ssl opts
												  #;
												  stmt
												  #;
												  prmbnds
												  #;
												  resbnds)
	mysql-connection?
	(host mysql-connection-host)
	(user mysql-connection-user)
	(passwd mysql-connection-passwd)
	(db mysql-connection-db)
	(port mysql-connection-port)
	(unix-socket mysql-connection-unix-socket)
	(client-flag mysql-connection-client-flag)
	(connptr mysql-connection-ptr mysql-connection-ptr-set!)
	(result mysql-connection-result mysql-connection-result-set!)
	(result-start mysql-connection-result-start mysql-connection-result-start-set!)
	(ssl mysql-connection-ssl)
	(opts mysql-connection-options)
	#;
	(stmt mysql-connection-statement mysql-connection-statement-set!)
	#; ; Toggles btwn param & result
	(bnds mysql-connection-bindings mysql-connection-bindings-set!) )

;;

(define (make-mysql-connection mysql host user passwd db port unix-socket client-flag options ssl)
  (mysql-ssl-set! mysql ssl)
  (mysql-options-set! mysql options)
  (let ([connptr (foreign-mysql-real-connect mysql
                                             host user passwd
                                             db
                                             port unix-socket
                                             client-flag)])
    (if mysqlptr
        (%make-mysql-connection host user passwd db port unix-socket
                                client-flag connptr #f #f options ssl
                                #;
                                #f
                                #;
                                #f)
        (signal-mysql-condition 'mysql-connect
         (foreign-mysql-error mysql)
         host user passwd db port unix-socket client-flag options ssl options) ) ) )

; MySQL Connection verification functions

(define (mysql-is-closed? conn)
  (not (mysql-connection-ptr conn)))
  
(define (mysql-check-connection conn)
  (if (mysql-is-closed? conn)
      (signal-mysql-condition 'mysql-check-connection "connection is closed")))

;; Specialized connection record printers

 ;DEBUG
(define-record-printer (mysql-connection conn out)
  (mysql-check-connection conn)
	(let ([host (mysql-connection-host conn)]
				[user (mysql-connection-user conn)]
				[passwd (mysql-connection-passwd conn)]
				[db (mysql-connection-db conn)]
				[tcp-port (mysql-connection-port conn)]
				[unix-socket (mysql-connection-unix-socket conn)]
				[client-flag (mysql-connection-client-flag conn)]
				[ssl (mysql-connection-ssl conn)]
				[opts (mysql-connection-options conn)]
				#;
				[stmt (mysql-connection-statement conn)]
				#;
				[bnds (mysql-connection-bindings conn)])
		(display
		 (string-append
			"#<mysql-connection"
			(if (mysql-connection-ptr conn)
					(string-append
						(optional-value->string host					"host")
						(optional-value->string user					"user")
						(optional-value->string passwd				"passwd")
						(optional-value->string db						"db")
						(optional-value->string tcp-port			"tcp-port"		(not (fx= 0 tcp-port)))
						(optional-value->string unix-socket   "unix-socket")
						(optional-value->string client-flag   "client-flag" (not (fx= 0 client-flag)))
						(optional-value->string ssl					  "ssl")
						(optional-value->string opts					"options")
						#;
						(optional-value->string stmt					"statement")
						#;
						(optional-value->string bnds		      "bindings") )
					" INVALID")
			">")
		 out) ) )

#; ;RELEASE
(define-record-printer (mysql-connection conn out)
  (mysql-check-connection conn)
	(let ([host (mysql-connection-host conn)]
				[user (mysql-connection-user conn)]
				[passwd (mysql-connection-passwd conn)]
				[db (mysql-connection-db conn)]
				[tcp-port (mysql-connection-port conn)]
				[unix-socket (mysql-connection-unix-socket conn)]
				[client-flag (mysql-connection-client-flag conn)]
				[ssl (mysql-connection-ssl conn)]
				[opts (mysql-connection-options conn)])
		(display
		 (string-append
			"#<mysql-connection"
			(if (mysql-connection-ptr conn)
					(string-append
						(optional-value->string host  "host")
						(optional-value->string user  "user") )
					" INVALID")
			">")
		 out) ) )

;-----------------------------------------------------------------------
; The "base" MySQL/Scheme API.
;
; This part of the API provides a slightly simplified version of the full
; MySQL C API.
;

;;

(define (mysql-affected-rows conn)
  (mysql-check-connection conn)
	(let ([cnt (foreign-mysql-affected-rows (mysql-connection-ptr conn))])
		(and (not (= -1 cnt))
				 cnt ) ) )

;;

(define (mysql-change-user conn #!key (user #f) (passwd #f) (db #f))
  (mysql-check-connection conn)
	(foreign-mysql-change-user (mysql-connection-ptr conn) user passwd db) )

;;

(define (mysql-character-set-name conn)
  (mysql-check-connection conn)
	(foreign-mysql-character-set-name (mysql-connection-ptr conn)))

;; Closes a mysql connection and invalidates the mysql connection object.
;; Returns (void). You should do this when you're done with the MySQL
;; connection; however, if you don't close it manually, it will be closed
;; upon termination.

(define (mysql-close conn)
	#;
	(mysql-stmt-close conn)
	(mysql-free-result conn)
	(foreign-mysql-close (mysql-connection-ptr conn))
	(mysql-connection-ptr-set! conn #f) )

;; Returns a mysql connection object, or #f on failure.

(define (mysql-connect #!key (host #f) (user #f) (passwd #f) (db #f) (port 0)
											 (unix-socket #f) (client-flag 0)
											 (options #f) (ssl #f))
	(let ([mysql (foreign-mysql-init #f)])
		(if mysql
        (make-mysql-connection mysql host user passwd db port unix-socket client-flag options ssl)
        (signal-mysql-condition 'mysql-connect "failed to initialize mysql") ) ) )

;;

(define (mysql-debug debug)
	(foreign-mysql-debug debug) )

;;

(define (mysql-dump-debug-info conn)
  (mysql-check-connection conn)
	(foreign-mysql-dump-debug-info (mysql-connection-ptr conn)) )

;;

(define (mysql-errno conn)
  (mysql-check-connection conn)
	(foreign-mysql-errno (mysql-connection-ptr conn)))

;; Returns a string describing the last mysql error, or #f if no error
;; has occurred.

(define (mysql-error conn)
  (mysql-check-connection conn)
	(let ([errstr (foreign-mysql-error (mysql-connection-ptr conn))])
		(and (not (string=? "" errstr))
				 errstr) ) )

;;

(define (mysql-escape-string conn str)
	(let-location ([escstr c-string*])
		((foreign-lambda* void ((mysql-ptr mysql) (c-pointer to) (c-string from) (unsigned-long length))
			"if ((*((char **) to) = ((char *) C_malloc( (2 * length) + 1 )))) {\n"
			"		 (void) mysql_real_escape_string( mysql, *((char **) to), from, length );\n"
			"}")
		 #$escstr
		 str (string-length str))
		escstr ) )

;; Returns a mysql-field-ptr or #f when no more fields.
;; Returns #f when no result set.

(define (mysql-fetch-field conn)
	(and-let* ([resptr (mysql-connection-result conn)])
		(foreign-mysql-fetch-field resptr) ) )

;; Returns a mysql-field-ptr to a vector of MYSQL_FIELD or #f when no fields.
;; Returns #f when no result set.

(define (mysql-fetch-fields conn)
	(and-let* ([resptr (mysql-connection-result conn)])
		(foreign-mysql-fetch-fields resptr) ) )

;; Returns a mysql-field-ptr or #f when no such field.
;; Returns #f when no result set.

(define (mysql-fetch-field-direct conn field-number)
	(and-let* ([resptr (mysql-connection-result conn)])
		(foreign-mysql-fetch-field-direct resptr field-number) ) )

;; Returns a u32vector of length num-fields.
;; Returns #f when no result set.

(define (mysql-fetch-lengths-internal resptr cnt)
	(and-let* ([ulptr (foreign-mysql-fetch-lengths resptr)])
		(unsigned-long-array->u32vector ulptr cnt) ) )

;; Returns a u32vector of length num-fields.
;; Returns #f when no result set.

(define (mysql-fetch-lengths conn)
	(and-let* ([resptr (mysql-connection-result conn)])
		(mysql-fetch-lengths-internal resptr (foreign-mysql-num-fields resptr)) ) )

;; After a mysql-query that has results, use mysql-fetch-row to retrieve
;; results row-by-row. When no more rows are left, returns #f. When returning
;; a "row", returns a procedure that takes exactly 1 argument, which may
;; be either a number (in which case it is treated as the column index,
;; starting at zero) or a symbol or string (which will be treated as the
;; column name).

(define (mysql-fetch-row conn)
	(and-let* ([resptr (mysql-connection-result conn)]
						 [row (foreign-mysql-fetch-row resptr)])
		(let ([fldcnt (foreign-mysql-num-fields resptr)]
					[fldlens #f])
			(lambda (field)
				(and-let* ([fldidx (mysql-get-field-index resptr field fldcnt)])
					(if (foreign-mysqlaux-is-binary-field resptr fldidx)
							(binary-char-pointer->string
							 (foreign-mysqlaux-fetch-column-data-direct row fldidx)
							 (u32vector-ref
								(or fldlens
										(begin
											(set! fldlens (mysql-fetch-lengths-internal resptr fldcnt))
											fldlens ) )
								fldidx))
							(foreign-mysqlaux-fetch-column-string-direct row fldidx) ) ) ) ) ) )

;;

(define (mysql-field-count conn)
  (mysql-check-connection conn)
	(foreign-mysql-field-count (mysql-connection-ptr conn)) )

;;

(define (mysql-free-result conn)
	(and-let* ([res (mysql-connection-result conn)])
		(foreign-mysql-free-result res) )
	(mysql-connection-result-set! conn #f)
	(mysql-connection-result-start-set! conn #f) )

;; Returns a c-pointer to a MY_CHARSET_INFO struct.
;; a finalizer is supplied.

(define (mysql-get-character-set-info conn)
  (mysql-check-connection conn)
	(let ([chrsetinfo (allocate-my-charset-info)])
		(foreign-mysql-get-character-set-info (mysql-connection-ptr conn) chrsetinfo)
		(set-finalizer! chrsetinfo free-my-charset-info)
		chrsetinfo ) )

;;

(define (mysql-get-client-info)
	(foreign-mysql-get-client-info) )

;;

(define (mysql-get-client-version)
	(foreign-mysql-get-client-version) )

;;

(define (mysql-get-host-info conn)
  (mysql-check-connection conn)
	(foreign-mysql-get-host-info (mysql-connection-ptr conn)) )

;;

(define (mysql-get-proto-info conn)
  (mysql-check-connection conn)
	(foreign-mysql-get-proto-info (mysql-connection-ptr conn)) )

;;

(define (mysql-get-server-info conn)
  (mysql-check-connection conn)
	(foreign-mysql-get-server-info (mysql-connection-ptr conn)) )

;;

(define (mysql-get-server-version conn)
  (mysql-check-connection conn)
	(foreign-mysql-get-server-version (mysql-connection-ptr conn)) )

;;

(define (mysql-info conn)
  (mysql-check-connection conn)
	(foreign-mysql-info (mysql-connection-ptr conn)) )

;;

(define (mysql-insert-id conn)
  (mysql-check-connection conn)
	(foreign-mysql-insert-id (mysql-connection-ptr conn)) )

;;

(define (mysql-kill conn pid)
  (mysql-check-connection conn)
	(foreign-mysql-kill (mysql-connection-ptr conn) pid) )

;;

(define (mysql-list-dbs conn like)
  (mysql-check-connection conn)
	(mysql-free-result conn)
	(mysql-connection-result-set! conn
		(foreign-mysql-list-dbs (mysql-connection-ptr conn) like)) )

;;

(define (mysql-list-fields conn table wild)
  (mysql-check-connection conn)
	(mysql-free-result conn)
	(mysql-connection-result-set! conn
		(foreign-mysql-list-fields (mysql-connection-ptr conn) table wild)) )

;;

(define (mysql-list-processes conn)
  (mysql-check-connection conn)
	(mysql-free-result conn)
	(mysql-connection-result-set! conn
		(foreign-mysql-list-processes (mysql-connection-ptr conn))) )

;;

(define (mysql-list-tables conn wild)
  (mysql-check-connection conn)
	(mysql-free-result conn)
	(mysql-connection-result-set! conn
		(foreign-mysql-list-tables (mysql-connection-ptr conn) wild)) )

;;

(define (mysql-num-fields conn)
	(foreign-mysql-num-fields (mysql-connection-result conn)) )

;;

(define (mysql-num-rows conn)
	(and-let* ([res (mysql-connection-result conn)])
		(foreign-mysql-num-rows res)) )

;;

(define (mysql-ping conn)
  (mysql-check-connection conn)
	(foreign-mysql-ping (mysql-connection-ptr conn)) )

;; Returns #t if the query was successful, signals exception otherwise.

(define (mysql-query conn query)
  (mysql-check-connection conn)
	(let ([mysql-ptr (mysql-connection-ptr conn)])
		; zero indicates success
		(if (fx= 0 (foreign-mysql-real-query mysql-ptr query (string-length query)))
				(begin (mysql-store-result conn) #t)
				(signal-mysql-error 'mysql-query conn query) ) ) )

;; Returns #t if the select was successful, signals exception otherwise.

(define (mysql-select-db conn db)
  (mysql-check-connection conn)
	(or (fx= 0 (foreign-mysql-select-db (mysql-connection-ptr conn) db))
			(signal-mysql-error 'mysql-select-db conn db) ) )

;; Returns #t if the set was successful, signals exception otherwise.

(define (mysql-set-character-set conn csname)
  (mysql-check-connection conn)
	(or (fx= 0 (foreign-mysql-set-character-set (mysql-connection-ptr conn) csname))
			(signal-mysql-condition 'mysql-set-character-set "bad character set name" csname) ) )

;;

(define (mysql-stat conn)
  (mysql-check-connection conn)
	(foreign-mysql-stat (mysql-connection-ptr conn)) )

;;

(define (mysql-store-result conn)
  (mysql-check-connection conn)
	(mysql-connection-result-set! conn
	 (foreign-mysql-store-result (mysql-connection-ptr conn)))
	(mysql-connection-result-start-set! conn
	(and (mysql-connection-result conn)
			 (foreign-mysql-row-tell (mysql-connection-result conn))) ) )

;;

(define (mysql-thread-id conn)
  (mysql-check-connection conn)
	(foreign-mysql-thread-id (mysql-connection-ptr conn)) )

;-----------------------------------------------------------------------
; The "extended" MySQL/Scheme API.
;
; This API provides some additional functionality.
;

;; Rewinds to the beginning of the result set. has no effect if there is no
;; current result set.

(define (mysql-rewind conn)
	(and-let* ([resptr (mysql-connection-result-start conn)])
		(foreign-mysql-row-seek (mysql-connection-result conn) resptr) ) )

;-----------------------------------------------------------------------
; The "map" MySQL/Scheme API.
;
; This API provides some additional functionality for traversing results
; in a Scheme-ish way.
;

;; calls proc on every row in the current result set. proc should take 3
;; arguments: the row (as described for mysql-fetch-row), the row index
;; (which starts with 1 and ends with (mysql-num-rows conn)), and the
;; current accumulated value.
;;
;; Returns the final accumulated value.
;;
;; note: rewinds the result set before and after iterating over it; thus,
;; all rows are included.
;;
;; you must call mysql-rewind if you later want to iterate over the result set
;; using mysql-fetch-row.

(define (mysql-row-fold conn proc init)
	(mysql-rewind conn)
	(let loop ([rownum 1] [acc init])
		(let ([row (mysql-fetch-row conn)])
			(if row
					(loop (fx+ rownum 1) (proc row rownum acc))
					acc ) ) ) )

;; calls proc on every row in the current result set. proc should take 2
;; arguments: the row (as described for mysql-fetch-row) and the row index
;; (which starts with 1 and ends with (mysql-num-rows conn)).
;;
;; note: rewinds the result set before and after iterating over it; thus,
;; all rows are included.
;;
;; you must call mysql-rewind if you later want to iterate over the result set
;; using mysql-fetch-row.

(define (mysql-row-for-each conn proc)
	(mysql-row-fold conn
									(lambda (row rownum _) (proc row rownum))
									#t) )

;; calls proc on every row in the current result set. proc should take 2
;; arguments: the row (as described for mysql-fetch-row) and the row index
;; (which starts with 1 and ends with (mysql-num-rows conn)).
;;
;; Returns a list of the results of each proc invocation.
;;
;; note: rewinds the result set before and after iterating over it; thus,
;; all rows are included.
;;
;; you must call mysql-rewind if you later want to iterate over the result set
;; using mysql-fetch-row.

(define (mysql-row-map conn proc)
	(reverse!
		(mysql-row-fold conn
										(lambda (row rownum lst) (cons (proc row rownum) lst))
										'())) )

;; executes query and then mysql-row-for-each with the given proc. The proc
;; must meet the contract specified for the proc passed to mysql-row-for-each.

(define (mysql-query-fold conn query proc init)
	(mysql-query conn query)
	(mysql-row-fold conn proc init) )

;; executes query and then mysql-row-for-each with the given proc. The proc
;; must meet the contract specified for the proc passed to mysql-row-for-each.

(define (mysql-query-for-each conn query proc)
	(mysql-query conn query)
	(mysql-row-for-each conn proc) )

;; executes query and then mysql-row-for-each with the given proc. The proc
;; must meet the contract specified for the proc passed to mysql-row-for-each.

(define (mysql-query-map conn query proc)
	(mysql-query conn query)
	(mysql-row-map conn proc) )

;; Synonyms

(define mysql-query-foreach mysql-query-for-each)
(define mysql-foreach-row mysql-row-for-each)

;-----------------------------------------------------------------------
; The MySQL Field structure predicate API.
;

;;

(define (mysql-field-flags-test fldptr mask)
	(bitwise-and (mysql-field-flags fldptr) mask) )

;;

(define (mysql-field-flags-mask flags)
	(apply bitwise-ior flags) )

;;

(define (mysql-field-flags-on? fldptr . flags)
	(let ([mask (mysql-field-flags-mask flags)])
		(= mask (mysql-field-flags-test fldptr mask)) ) )

;;

(define (mysql-field-flags-off? fldptr . flags)
	(fx= 0 (mysql-field-flags-test fldptr (mysql-field-flags-mask flags))) )

;;

(define (mysql-field-type-any? fldptr . types)
	(memv (mysql-field-type fldptr) types) )

;;

(define (mysql-field-type=? fldptr type)
	(eqv? type (mysql-field-type fldptr)) )

;;

(define (mysql-field-primary-key? fldptr)
	(mysql-field-flags-on? fldptr pri-key-flag) )

;;

(define (mysql-field-not-null? fldptr)
	(mysql-field-flags-on? fldptr not-null-flag) )

;;

(define (mysql-field-binary? fldptr)
	(= 63 (mysql-field-charsetnr fldptr)) )

;;

(define (mysql-field-numeric? fldptr)
	(mysql-field-flags-on? fldptr num-flag) )

;;

(define mysql-field-type-clock?
	(let ([numtypes (list mysql-type-timestamp mysql-type-datetime
												mysql-type-date mysql-type-time
												mysql-type-newdate mysql-type-year)])
		(lambda (fldptr)
			(apply mysql-field-type-any? fldptr numtypes) ) ) )

;; note - not the same as the "IS_NUM" macro.

(define mysql-field-type-number?
	(let ([numtypes (list mysql-type-decimal mysql-type-tiny mysql-type-short
												mysql-type-long mysql-type-float mysql-type-double
												mysql-type-longlong mysql-type-newdecimal
												mysql-type-bit)])
		(lambda (fldptr)
			(apply mysql-field-type-any? fldptr numtypes) ) ) )

;;

(define mysql-field-type-blob?
	(let ([blobtypes (list mysql-type-tiny-blob mysql-type-medium-blob
												 mysql-type-long-blob mysql-type-blob)])
		(lambda (fldptr)
			(apply mysql-field-type-any? fldptr blobtypes) ) ) )

;;

(define mysql-field-type-string?
	(let ([numtypes (list mysql-type-varchar mysql-type-var-string
												mysql-type-string
												mysql-type-enum mysql-type-set)])
		(lambda (fldptr)
			(apply mysql-field-type-any? fldptr numtypes) ) ) )

;;
;; note - the same as the "IS_NUM" macro.

(define mysql-field-type-magnitude?
	(let ([magtypes (list mysql-type-timestamp mysql-type-year mysql-type-null)])
		(lambda (fldptr)
			(or (mysql-field-type-number? fldptr)
					(apply mysql-field-type-any? fldptr magtypes) ) ) ) )

;;

(define (mysql-field-type-binary? fldptr)
	(and (mysql-field-binary? fldptr)
			 (or (mysql-field-type-blob? fldptr)
					 (mysql-field-type-string? fldptr) ) ) )

;;

(define (mysql-field-type-text? fldptr)
	(and (not (mysql-field-binary? fldptr))
			 (or (mysql-field-type-blob? fldptr)
					 (mysql-field-type-string? fldptr) ) ) )

;-----------------------------------------------------------------------
; The MySQL Field structure multi-slot API.
;

;; Returns a list of field items.

(define (mysql-field-slots fldptr . getters)
	(and fldptr
			 (map (cut <> fldptr) getters) ) )

;; Returns a list of field items for nth field.

(define (mysql-fetch-field-slots-direct conn nth . getters)
	(apply mysql-field-slots (mysql-fetch-field-direct conn nth) getters) )

;; Returns a field item for nth field.

(define (mysql-fetch-field-slot-direct conn nth getter)
	(and-let* ([lst (mysql-field-slots (mysql-fetch-field-direct conn nth) getter)])
		(car lst) ) )

;; Returns a list of field items for the next field.

(define (mysql-fetch-field-slots conn . getters)
	(apply mysql-field-slots (mysql-fetch-field conn) getters) )

;; Returns a field item for the next field.

(define (mysql-fetch-field-slot conn getter)
	(and-let* ([lst (mysql-field-slots (mysql-fetch-field conn) getter)])
		(car lst) ) )

;; Returns a field pointer or #f.

(define (mysql-fetch-field-specific conn field)
	(and-let* ([resptr (mysql-connection-result conn)]
						 [fldidx (mysql-get-field-index resptr field (foreign-mysql-num-fields resptr))])
		(foreign-mysql-fetch-field-direct resptr fldidx) ) )

#|
;=======================================================================
; The MYSQL_TIME API.
;

(declare
	(export
		;; enum enum_mysql_timestamp_type
		mysql-timestamp-date
		mysql-timestamp-datetime
		mysql-timestamp-error
		mysql-timestamp-none
		mysql-timestamp-time
		;
		mysql-timestamp-type-symbol
		mysql-timestamp-type-value
		;; MYSQL_TIME
		mysql-time-day-set!
		mysql-time-hour-set!
		mysql-time-minute-set!
		mysql-time-month-set!
		mysql-time-neg-set!
		mysql-time-second-part-set!
		mysql-time-second-set!
		mysql-time-time-type-set!
		mysql-time-year-set!
		;
		mysql-time-day
		mysql-time-hour
		mysql-time-minute
		mysql-time-month
		mysql-time-neg
		mysql-time-second
		mysql-time-second-part
		mysql-time-time-type
		mysql-time-year
		;
		make-mysql-time
		allocate-mysql-time
		free-mysql-time ) )

;;

(define-foreign-type mysql-time-ptr (c-pointer "MYSQL_TIME"))

;;

($define-foreign-enum (mysql-timestamp-type (enum "enum_mysql_timestamp_type"))
	#f	; No aliases!
	MYSQL_TIMESTAMP_NONE
	MYSQL_TIMESTAMP_ERROR
	MYSQL_TIMESTAMP_DATE
	MYSQL_TIMESTAMP_DATETIME
	MYSQL_TIMESTAMP_TIME )

;;

(define-foreign-record (mysql-time "MYSQL_TIME")
	(rename: c-name->scheme-name)
	(constructor: allocate-mysql-time)
	(destructor: free-mysql-time)
	(unsigned-int year)
	(unsigned-int month)
	(unsigned-int day)
	(unsigned-int hour)
	(unsigned-int minute)
	(unsigned-int second)
	(unsigned-long second_part)
	(my-bool neg)
	(mysql-timestamp-type time_type) )

;;

(define (make-mysql-time type #!key (year 0) (month 0) (day 0)
																		(hour 0) (minute 0) (second 0)
																		(second-part 0)
																		is-negative)
	(let ([timptr (allocate-mysql-time)])
		(mysql-time-time-type-set! timptr type)
		(mysql-time-year-set! timptr year)
		(mysql-time-month-set! timptr month)
		(mysql-time-day-set! timptr day)
		(mysql-time-hour-set! timptr hour)
		(mysql-time-minute-set! timptr minute)
		(mysql-time-second-set! timptr second)
		(mysql-time-second-part-set! timptr second-part)
		(mysql-time-neg-set! timptr is-negative)
		(set-finalizer! timptr free-mysql-time)
		timptr ) )

;; #(seconds minutes hours mday month year wday yday dstflag timezone)

(define (mysql-time->time-vector timptr)
  (let ([timvec (make-vector 10 #f)])
    (cond [(or (eq? mysql-timestamp-datetime (mysql-timestamp-type timptr))
               (eq? mysql-timestamp-time (mysql-timestamp-type timptr)))
            (vector-set! timvec 0 (mysql-time-second timptr))
            (vector-set! timvec 1 (mysql-time-minute timptr))
            (vector-set! timvec 2 (mysql-time-hour timptr))]
          [else
            (vector-set! timvec 0 #f)
            (vector-set! timvec 1 #f)
            (vector-set! timvec 2 #f)])
    (cond [(or (eq? mysql-timestamp-datetime (mysql-timestamp-type timptr))
               (eq? mysql-timestamp-date (mysql-timestamp-type timptr)))
            (vector-set! timvec 3 (mysql-time-day timptr))
            (vector-set! timvec 4 (mysql-time-month timptr))
            (vector-set! timvec 5 (mysql-time-year timptr))]
          [else
            (vector-set! timvec 3 #f)
            (vector-set! timvec 4 #f)
            (vector-set! timvec 5 #f)])
    (vector-set! timvec 6 #f)
    (vector-set! timvec 7 #f)
    (vector-set! timvec 8 #f)
    (vector-set! timvec 9 #f)
    timvec ) )

;; #(seconds minutes hours mday month year wday yday dstflag timezone)

(define (time-vector->mysql-time timvec #!optional (timtyp mysql-timestamp-datetime))
  (let ([timptr (make-mysql-time timtyp)])
    (mysql-time-second-set! timptr (vector-ref timvec 0))
    (mysql-time-minute-set! timptr (vector-ref timvec 1))
    (mysql-time-hour-set! timptr (vector-ref timvec 2))
    (mysql-time-day-set! timptr (vector-ref timvec 3))
    (mysql-time-month-set! timptr (vector-ref timvec 4))
    (mysql-time-year-set! timptr (vector-ref timvec 5))
    (mysql-timestamp-type-set! timptr timtyp)
    timptr ) )

;=======================================================================
; The MySQL prepared statement API.
;

(declare
	(bound-to-procedure
		mysql-stmt-errno
		mysql-stmt-error
		mysql-stmt-sqlstate )
	(export
		;;
		stmt-attr-cursor-type
		stmt-attr-prefetch-rows
		stmt-attr-update-max-length
		;
		mysql-stmt-attr-type-symbol
		mysql-stmt-attr-type-value
		;;
		mysql-stmt-init-done
		mysql-stmt-execute-done
		mysql-stmt-prepare-done
		mysql-stmt-fetch-done
		;
		mysql-stmt-state-symbol
		mysql-stmt-state-value
		;;
		mysql-no-data
		mysql-data-truncated
		;
		mysql-status-return-code-symbol
		mysql-status-return-code-value
		;;
		foreign-mysql-stmt-affected-rows
		foreign-mysql-stmt-attr-get
		foreign-mysql-stmt-attr-set
		foreign-mysql-stmt-bind-param
		foreign-mysql-stmt-bind-result
		foreign-mysql-stmt-close
		foreign-mysql-stmt-data-seek
		foreign-mysql-stmt-errno
		foreign-mysql-stmt-error
		foreign-mysql-stmt-execute
		foreign-mysql-stmt-fetch
		foreign-mysql-stmt-fetch-column
		foreign-mysql-stmt-field-count
		foreign-mysql-stmt-free-result
		foreign-mysql-stmt-init
		foreign-mysql-stmt-insert-id
		foreign-mysql-stmt-num-rows
		foreign-mysql-stmt-param-count
		foreign-mysql-stmt-param-metadata
		foreign-mysql-stmt-prepare
		foreign-mysql-stmt-reset
		foreign-mysql-stmt-result-metadata
		foreign-mysql-stmt-row-seek
		foreign-mysql-stmt-row-tell
		foreign-mysql-stmt-send-long-data
		foreign-mysql-stmt-sqlstate
		foreign-mysql-stmt-store-result
		;; basic
		mysql-stmt-errno
		mysql-stmt-error
		mysql-stmt-sqlstate
		mysql-stmt-init ; custom
		mysql-stmt-close ; called by custom mysql-stmt-init
		mysql-stmt-prepare
		mysql-stmt-param-count
		mysql-stmt-bind-param
		mysql-stmt-execute
		mysql-stmt-affected-rows
		mysql-stmt-bind-result
		mysql-stmt-fetch
		mysql-stmt-store-result
		mysql-stmt-result-metadata
		mysql-stmt-attr-set
		mysql-stmt-attr-get
		;; extended
		mysql-stmt-rewind
		mysql-stmt-row-fetch
		mysql-stmt-query
		;; mapping
		mysql-stmt-row-fold
		mysql-stmt-row-for-each
		mysql-stmt-row-map
		mysql-stmt-query-fold
		mysql-stmt-query-for-each
		mysql-stmt-query-map
		;; MYSQL_BIND
		allocate-mysql-bind
		free-mysql-bind
		mysql-bind-ref-ptr
		mysql-bind-clear!
		mysql-bind-param-init-direct
		mysql-bind-result-init-direct
		mysql-bind-param-init
		mysql-bind-result-init
		;
		mysql-bind-buffer-set!
		mysql-bind-buffer-length-set!
		mysql-bind-buffer-type-set!
		mysql-bind-error-set!
		mysql-bind-error-value-set!
		mysql-bind-is-null-set!
		mysql-bind-is-null-value-set!
		mysql-bind-is-unsigned-set!
		mysql-bind-length-set!
		;
		mysql-bind-buffer
		mysql-bind-buffer-length
		mysql-bind-buffer-type
		mysql-bind-error
		mysql-bind-error-value
		mysql-bind-is-null
		mysql-bind-is-null-value
		mysql-bind-is-unsigned
		mysql-bind-length ) )

;;

(define-foreign-type mysql-row-offset mysql-rows-ptr)

(define-foreign-type mysql-bind-ptr (c-pointer "MYSQL_BIND"))

(define-foreign-type mysql-stmt-ptr (c-pointer "MYSQL_STMT"))

(define-foreign-type mysql-my-bool-ptr (c-pointer "my_bool"))

(define-foreign-type mysql-ulong-ptr (c-pointer "unsigned long"))

;;

($define-foreign-enum (mysql-stmt-state (enum "enum_mysql_stmt_state"))
	#f	; No aliases!
	MYSQL_STMT_INIT_DONE
	MYSQL_STMT_PREPARE_DONE
	MYSQL_STMT_EXECUTE_DONE
	MYSQL_STMT_FETCH_DONE )

($define-foreign-enum (mysql-stmt-attr-type (enum "enum_stmt_attr_type"))
	#f	; No aliases!
	STMT_ATTR_UPDATE_MAX_LENGTH
	STMT_ATTR_CURSOR_TYPE
	STMT_ATTR_PREFETCH_ROWS )

($define-foreign-enum (mysql-status-return-code unsigned-int)
	#f	; No aliases!
	MYSQL_NO_DATA
	MYSQL_DATA_TRUNCATED )

;;

;MYSQL_STMT *mysql_stmt_init(MYSQL *mysql)
(define foreign-mysql-stmt-init
	(foreign-lambda mysql-stmt-ptr "mysql_stmt_init" mysql-ptr))

;my_bool mysql_stmt_close(MYSQL_STMT *mysql)
(define foreign-mysql-stmt-close
	(foreign-lambda my-bool "mysql_stmt_close" mysql-stmt-ptr))

;my_bool mysql_stmt_bind_param(MYSQL_STMT *stmt, MYSQL_BIND *bind)
(define foreign-mysql-stmt-bind-param
	(foreign-lambda my-bool "mysql_stmt_bind_param" mysql-stmt-ptr mysql-bind-ptr))

;my_bool mysql_stmt_bind_result(MYSQL_STMT *stmt, MYSQL_BIND *bind)
(define foreign-mysql-stmt-bind-result
	(foreign-lambda my-bool "mysql_stmt_bind_result" mysql-stmt-ptr mysql-bind-ptr))

;int mysql_stmt_execute(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-execute
	(foreign-lambda int "mysql_stmt_execute" mysql-stmt-ptr))

;int mysql_stmt_fetch(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-fetch
	(foreign-lambda int "mysql_stmt_fetch" mysql-stmt-ptr))

;int mysql_stmt_prepare(MYSQL_STMT *stmt, const char *query, unsigned long length)
(define foreign-mysql-stmt-prepare
	(foreign-lambda int "mysql_stmt_prepare" mysql-stmt-ptr c-string unsigned-long))

;int mysql_stmt_fetch_column(MYSQL_STMT *stmt, MYSQL_BIND *bind_arg, unsigned int column, unsigned long offset)
(define foreign-mysql-stmt-fetch-column
	(foreign-lambda int "mysql_stmt_fetch_column" mysql-stmt-ptr mysql-bind-ptr unsigned-int unsigned-long))

;int mysql_stmt_store_result(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-store-result
	(foreign-lambda int "mysql_stmt_store_result" mysql-stmt-ptr))

;unsigned long mysql_stmt_param_count(MYSQL_STMT * stmt)
(define foreign-mysql-stmt-param-count
	(foreign-lambda unsigned-long "mysql_stmt_param_count" mysql-stmt-ptr))

;my_bool mysql_stmt_attr_set(MYSQL_STMT *stmt, enum enum_stmt_attr_type attr_type, const void *attr)
(define foreign-mysql-stmt-attr-set
	(foreign-lambda my-bool "mysql_stmt_attr_set" mysql-stmt-ptr mysql-stmt-attr-type c-pointer))

;my_bool mysql_stmt_attr_get(MYSQL_STMT *stmt, enum enum_stmt_attr_type attr_type, void *attr)
(define foreign-mysql-stmt-attr-get
	(foreign-lambda my-bool "mysql_stmt_attr_get" mysql-stmt-ptr mysql-stmt-attr-type c-pointer))

(define foreign-mysqlaux-stmt-attr-set-bool
	(foreign-lambda* my-bool ((mysql-stmt-ptr stmt) (mysql-stmt-attr-type attrr) (my-bool value))
   "return( mysql_stmt_attr_set( stmt, attr, &value ) );"))

(define foreign-mysqlaux-stmt-attr-set-ulong
	(foreign-lambda* my-bool ((mysql-stmt-ptr stmt) (mysql-stmt-attr-type attrr) (unsigned-long value))
   "return( mysql_stmt_attr_set( stmt, attr, &value ) );"))

;my_bool mysql_stmt_reset(MYSQL_STMT * stmt)
(define foreign-mysql-stmt-reset
	(foreign-lambda my-bool "mysql_stmt_reset" mysql-stmt-ptr))

;my_bool mysql_stmt_free_result(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-free-result
	(foreign-lambda my-bool "mysql_stmt_free_result" mysql-stmt-ptr))

;my_bool mysql_stmt_send_long_data(MYSQL_STMT *stmt, unsigned-int param_number, const char *data, unsigned long length)
(define foreign-mysql-stmt-send-long-data
	(foreign-lambda my-bool "mysql_stmt_send_long_data" mysql-stmt-ptr unsigned-int c-pointer unsigned-long))

;MYSQL_RES *mysql_stmt_result_metadata(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-result-metadata
	(foreign-lambda mysql-res-ptr "mysql_stmt_result_metadata" mysql-stmt-ptr))

;MYSQL_RES *mysql_stmt_param_metadata(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-param-metadata
	(foreign-lambda mysql-res-ptr "mysql_stmt_param_metadata" mysql-stmt-ptr))

;unsigned int mysql_stmt_errno(MYSQL_STMT * stmt)
(define foreign-mysql-stmt-errno
	(foreign-lambda unsigned-int "mysql_stmt_errno" mysql-stmt-ptr))

;const char *mysql_stmt_error(MYSQL_STMT * stmt)
(define foreign-mysql-stmt-error
	(foreign-lambda c-string "mysql_stmt_error" mysql-stmt-ptr))

;const char *mysql_stmt_sqlstate(MYSQL_STMT * stmt)
(define foreign-mysql-stmt-sqlstate
	(foreign-lambda c-string "mysql_stmt_sqlstate" mysql-stmt-ptr))

;MYSQL_ROW_OFFSET mysql_stmt_row_seek(MYSQL_STMT *stmt, MYSQL_ROW_OFFSET offset)
(define foreign-mysql-stmt-row-seek
	(foreign-lambda mysql-row-offset "mysql_stmt_row_seek" mysql-stmt-ptr mysql-row-offset))

;MYSQL_ROW_OFFSET mysql_stmt_row_tell(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-row-tell
	(foreign-lambda mysql-row-offset "mysql_stmt_row_tell" mysql-stmt-ptr))

;void mysql_stmt_data_seek(MYSQL_STMT *stmt, my_ulonglong offset)
(define foreign-mysql-stmt-data-seek
	(foreign-lambda* void ((mysql-stmt-ptr stmt) (my-ulonglong offset))
   "mysql_stmt_data_seek( stmt, offset );"))

;my_ulonglong mysql_stmt_num_rows(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-num-rows
	(foreign-lambda* my-ulonglong ((mysql-stmt-ptr stmt))
   "return( (double) mysql_stmt_num_rows( stmt ) );"))

;my_ulonglong mysql_stmt_affected_rows(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-affected-rows
	(foreign-lambda* my-ulonglong ((mysql-stmt-ptr stmt))
   "return( (double) mysql_stmt_affected_rows( stmt ) );"))

;my_ulonglong mysql_stmt_insert_id(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-insert-id
	(foreign-lambda* my-ulonglong ((mysql-stmt-ptr stmt))
   "return( (double) mysql_stmt_insert_id( stmt ) );"))

;unsigned int mysql_stmt_field_count(MYSQL_STMT *stmt)
(define foreign-mysql-stmt-field-count
	(foreign-lambda unsigned-int "mysql_stmt_field_count" mysql-stmt-ptr))

;; MYSQL_BIND

(define-foreign-record (mysql-bind "MYSQL_BIND")
	(rename: c-name->scheme-name)
	; special ctor
	(destructor: free-mysql-bind)
	((c-pointer "unsigned long") length)		; output length pointer
	((c-pointer "my_bool") is_null)					; Pointer to null indicator
	(c-pointer buffer)											; buffer to get/put data
	((c-pointer "my_bool") error)						; set this if you want to track data truncations happened during fetch
	(unsigned-long buffer_length)						; output buffer length, must be set when fetching str/binary
	(mysql-type buffer_type)								; buffer type
	(my-bool is_unsigned)										; set if integer type is unsigned
	; ???
	(my-bool error_value)										; used if error is 0
	(my-bool is_null_value) )								; Used if is_null is 0

;; Returns a C-vector of MYSQL_BIND
;; Count must be at least 1!

(define (allocate-mysql-bind #!optional (cnt 1))
  (and (fx<= 1 cnt)
       (allocate (fx* cnt (foreign-value "sizeof( MYSQL_BIND )" int))) ) )

;;

(define (make-mysql-bind #!optional (cnt 1))
  (and-let* ([binds (allocate-mysql-bind cnt)])
    (set-finalizer! binds free-mysql-bind)
    binds ) )

;; Returns c-pointer to MYSQL_BIND at index in a C-vector of MYSQL_BIND
;; No range checks!

(define (mysql-bind-ref-ptr bndptr idx)
	((foreign-lambda* mysql-bind-ptr ((mysql-bind-ptr ptr) (unsigned-integer idx))
	 "return( &(ptr[ idx ]) );")
	 bndptr idx) )

;; Zeros MYSQL_BIND

(define (mysql-bind-clear! bndptr #!optional (idx 0) (cnt 1))
	((foreign-lambda* void ((mysql-bind-ptr ptr) (unsigned-integer idx) (unsigned-integer cnt))
	 "memset( &(ptr[ idx ]), 0, sizeof( MYSQL_BIND ) * cnt );")
	 bndptr idx cnt) )

;; Sets a MYSQL_BIND variable pointers

(define (mysql-bind-variables-set! bndptr nulptr errptr lenptr bufptr)
  ((foreign-lambda* void ((mysql-bind-ptr pbnd)
                          (mysql-my-bool-ptr pnul)
                          (mysql-my-bool-ptr perr)
                          (mysql-ulong-ptr plen)
                          (c-pointer pbuf))
#<<EOS
  pbnd->buffer_length = *plen; /* Do anyway, even for results */
  pbnd->buffer = pbuf
  pbnd->is_null = pnul;
  pbnd->length = plen;
  pbnd->error = perr;
EOS
   )
   bndptr nulptr errptr lenptr bufptr) )

;; Returns a C-vector of my_bool
;; Count must be at least 1!

(define (allocate-mysql-my-bool #!optional (cnt 1))
  (and (<= 1 cnt)
       (allocate (* cnt (foreign-value "sizeof( my_bool )" int))) ) )

;; Returns boolptr at index in a C-vector of my_bool
;; No range checks!

(define (mysql-my-bool-ref-ptr myboolptr idx)
	((foreign-lambda* mysql-my-bool-ptr ((mysql-my-bool-ptr ptr) (unsigned-integer idx))
	 "return( &(ptr[ idx ]) );")
	 myboolptr idx) )

;; Returns bool at index in a C-vector of my_bool
;; No range checks!

(define (mysql-my-bool-ref myboolptr idx)
	((foreign-lambda* my-bool ((mysql-my-bool-ptr ptr) (unsigned-integer idx))
	 "return( ptr[ idx ] );")
	 myboolptr idx) )

;; Sets bool at index in a C-vector of my_bool
;; No range checks!

(define (mysql-my-bool-set! myboolptr idx val)
	((foreign-lambda* my-bool ((mysql-my-bool-ptr ptr) (unsigned-integer idx) (my-bool val))
	 "ptr[ idx ] = val;")
	 myboolptr idx val) )

;; Zeros my-bool

(define (mysql-my-bool-clear! myboolptr #!optional (idx 0) (cnt 1))
	((foreign-lambda* void ((mysql-my-bool-ptr ptr) (unsigned-integer idx))
	 "memset( &(ptr[ idx ]), 0, sizeof( my_bool ) * cnt );")
	 myboolptr idx) )

;; Returns a C-vector of ulong
;; Count must be at least 1!

(define (allocate-mysql-ulong #!optional (cnt 1))
  (and (<= 1 cnt)
       (allocate (* cnt (foreign-value "sizeof( unsigned long )" int))) ) )

;; Returns boolptr at index in a C-vector of my_bool
;; No range checks!

(define (mysql-ulong-ref-ptr ulongptr idx)
	((foreign-lambda* mysql-ulong-ptr ((mysql-ulong-ptr ptr) (unsigned-integer idx))
	 "return( &(ptr[ idx ]) );")
	 ulongptr idx) )

;; Returns ulong at index in a C-vector of ulong
;; No range checks!

(define (mysql-ulong-ref ulongptr idx)
	((foreign-lambda* unsigned-long ((mysql-ulong-ptr ptr) (unsigned-integer idx))
	 "return( ptr[ idx ] );")
	 ulongptr idx) )

;; Sets ulong at index in a C-vector of ulong
;; No range checks!

(define (mysql-ulong-set! ulongptr idx val)
	((foreign-lambda* my-bool ((mysql-ulongptr-ptr ptr) (unsigned-integer idx) (unsigned-long val))
	 "ptr[ idx ] = val;")
	 ulongptr idx val) )

;; Zeros ulong

(define (mysql-ulong-clear! ulongptr #!optional (idx 0) (cnt 1))
	((foreign-lambda* void ((mysql-ulong-ptr ptr) (unsigned-integer idx) (unsigned-integer cnt))
	 "memset( &(ptr[ idx ]), 0, sizeof( unsigned long ) * cnt );")
	 ulongptr idx cnt) )

;; Returns pointer @ ptr + off

(define (mysql-char-ref-ptr ptr off)
	((foreign-lambda* c-pointer ((c-pointer ptr) (unsigned-integer off))
	 "return( &(((char *) ptr)[ off ]) );")
	 ptr off) )

;;

(define-inline (*clear-memory! ptr len)
	(when (and ptr (not (fx= 0 len)))
	  ((foreign-lambda* void ((c-pointer ptr) (unsigned-integer len))
     "memset( ptr, 0, len );")
     ptr len) ) )

;;

(define-inline (*allocate len)
  (and (not (fx= 0 len))
       (allocate len) ) )

;;

(define-inline (*free ptr)
  (when ptr (free ptr) ) )

;; Record holds binding storage pointers

(define-record-type mysql-binding
  (%make-mysql-binding cnt bnds nuls errs lens bufs)
  mysql-binding?
  (cnt    mysql-binding-count       #;mysql-binding-counts-set!)
  (bnds   mysql-binding-binds       #;mysql-binding-binds-set!)
  (nuls   mysql-binding-is-nulls    #;mysql-binding-is-nulls-set!)
  (errs   mysql-binding-errors      #;mysql-binding-errors-set!)
  (lens   mysql-binding-lengths     #;mysql-binding-lengths-set!)
  (bufs   mysql-binding-buffers     #;mysql-binding-buffers-set!) )

;; Free all variable storage

(define (free-mysql-binding-variables bindings)
  (when bindings
    (*free (mysql-binding-binds bindings))
    (*free (mysql-binding-is-nulls bindings))
    (*free (mysql-binding-errors bindings))
    (*free (mysql-binding-lengths bindings))
    (*free (mysql-binding-buffers bindings)) ) )

;; Allocates zero'ed storage for binding
;; Returns mysql-binding instance or #f
;; When fldcnt is 0 no allocation performed
;; When buftot is 0 no buffer allocation performed

(define (make-mysql-binding-direct fldcnt #!optional (buftot 0))
  (and (fx> 0 fldcnt)
       ;FIXME should signal out of memory
       (let ([bnds (allocate-mysql-bind fldcnt)]
             [nuls (allocate-mysql-my-bool fldcnt)]
             [errs (allocate-mysql-my-bool fldcnt)]
             [lens (allocate-mysql-ulong fldcnt)]
             [bufs (*allocate buftot)])
         (mysql-bind-clear! bnds 0 fldcnt)
         (mysql-my-bool-clear! nuls 0 fldcnt)
         (mysql-my-bool-clear! errs 0 fldcnt)
         (mysql-ulong-clear! lens 0 fldcnt)
         (*clear-memory! bufs buftot)
         (let ([bindings (%make-mysql-binding fldcnt bnds nuls errs lens bufs)])
           #;(set-finalizer! bindings free-mysql-binding-variables)
           bindings ) ) ) )

;; Returns the cummulative buffer length upto the supplied index

(define (mysql-binding-buffer-offset idx lens)
  (do ([i 0 (fx+ i 1)]
       [off 0 (fx+ off (mysql-ulong-ref lens i))])
      [(fx= i idx) off]) )

;; Sets the MYSQL_BIND variable pointers for the binding @ index
;; idx  : which binding

(define (mysql-binding-direct-set! idx bnds nuls errs lens bufs)
  (mysql-bind-variables-set!
   (mysql-bind-ref-ptr bnds idx)
   (mysql-my-bool-ref-ptr nuls idx)
   (mysql-my-bool-ref-ptr errs idx)
   (mysql-ulong-ref-ptr lens idx)
   (and bufs
        (mysql-char-ref-ptr bufs (mysql-binding-buffer-offset idx lens)))) )

;; Returns mysql-binding instance with allocated variables
;; lendef : list of buffer lengths or field count
;; When lendef is a count no buffer allocation is performed

(define (make-mysql-binding lendef)
  (if (number? lendef)
    (make-mysql-binding-direct lendef 0)
    (let ([fldcnt (length lendef)]
          [buftot (fold + 0 lendef)])
      (let ([bindings (make-mysql-binding-direct fldcnt buftot)])
        (let ([bnds (mysql-binding-binds bindings)]
              [nuls (mysql-binding-is-nulls bindings)]
              [errs (mysql-binding-errors bindings)]
              [lens (mysql-binding-lengths bindings)]
              [bufs (mysql-binding-buffers bindings)])
          (do ([idx 0 (fx+ idx 1)]
               [lst lendef (cdr lst)])
              [(fx= fldcnt idx) bindings]
            (mysql-binding-direct-set! idx bnds nuls errs lens bufs)
            (mysql-ulong-set! lens idx (car lst))) ) ) ) ) )

;;;; TODO
;
; - MySQL Type
;
;   - mysql-type-* & optional length
;
; - Parameters
;
;   - build 'mysql-binding' for a given set of Scheme objects
;   - need 'mysql-type'
;     - use asserted 'mysql-type', if any
;     - when "metadata" then can deduce 'mysql-type'
;     - otherwise deduce 'mysql-type' from Scheme type of Scheme object
;   - given 'mysql-type' & Scheme object can create C-type storage requirements
;
; - Results
;
;   - build 'mysql-binding' for a given set of 'mysql-type'
;   - need 'mysql-type'
;     - use asserted 'mysql-type', if any
;     - when "metadata" then can deduce 'mysql-type'
;     - otherwise deduce 'mysql-type' from Scheme type
;   - given 'mysql-type' & Scheme type can create C-type storage requirements

;;

>#
typedef union {
  MYSQL_TIME              t;
  char                    sc;
  unsigned char           uc;
  short int               ssi;
  unsigned short int      usi;
  int                     si;
  unsigned int            ui;
  long long int           slli;
  unsigned long long int  ulli;
  float                   f;
  double                  d;
  char                    cv[1];
} MYSQLAUX_BIND_VARIABLE_BUFFER;

/* Setter */

static void
set_bind_variable_t( MYSQLAUX_BIND_VARIABLE *var, MYSQL_TIME *val )
{
  var->buffer.t = val;
}

static void
set_bind_variable_sc( MYSQLAUX_BIND_VARIABLE *var, int val )
{
  var->buffer.sc = val;
}

static void
set_bind_variable_uc( MYSQLAUX_BIND_VARIABLE *var, unsigned int val )
{
  var->buffer.uc = val;
}

static void
set_bind_variable_si( MYSQLAUX_BIND_VARIABLE *var, int val )
{
  var->buffer.ssi = val;
}

static void
set_bind_variable_usi( MYSQLAUX_BIND_VARIABLE *var, unsigned int val )
{
  var->buffer.usi = val;
}

static void
set_bind_variable_i( MYSQLAUX_BIND_VARIABLE *var, int val )
{
  var->buffer.si = val;
}

static void
set_bind_variable_ui( MYSQLAUX_BIND_VARIABLE *var, unsigned int val )
{
  var->buffer.ui = val;
}

static void
set_bind_variable_lli( MYSQLAUX_BIND_VARIABLE *var,  val )
{
  var->buffer.lli = val;
}

static void
set_bind_variable_ulli( MYSQLAUX_BIND_VARIABLE *var, double val )
{
  var->buffer.ulli = val;
}

static void
set_bind_variable_f( MYSQLAUX_BIND_VARIABLE *var, double val )
{
  var->buffer.f = val;
}

static void
set_bind_variable_d( MYSQLAUX_BIND_VARIABLE *var, double val )
{
  var->buffer.d = val;
}

static void
set_bind_variable_cv( MYSQLAUX_BIND_VARIABLE *var, char *cv )
{
  memcpy( var->buffer.cv, cv, var->length );
}

/* Getter */

static MYSQL_TIME *
get_bind_variable_t( MYSQLAUX_BIND_VARIABLE *var )
{
  return &(var->buffer.t);
}

static int
get_bind_variable_sc( MYSQLAUX_BIND_VARIABLE *var )
{
  return var->buffer.sc;
}

static unsigned int
get_bind_variable_uc( MYSQLAUX_BIND_VARIABLE *var )
{
  return var->buffer.uc;
}

static int
get_bind_variable_ssi( MYSQLAUX_BIND_VARIABLE *var )
{
  return (int) var->buffer.ssi;
}

static unsigned int
get_bind_variable_usi( MYSQLAUX_BIND_VARIABLE *var )
{
  return var->buffer.usi;
}

static int
get_bind_variable_si( MYSQLAUX_BIND_VARIABLE *var )
{
  return (int) var->buffer.si;
}

static unsigned int
get_bind_variable_ui( MYSQLAUX_BIND_VARIABLE *var )
{
  return var->buffer.ui;
}

/*
  long long conversion does not preserve precision
  need to use Chicken Scheme bigint
*/

static double
get_bind_variable_slli( MYSQLAUX_BIND_VARIABLE *var )
{
  return (int) var->buffer.slli;
}

static double
get_bind_variable_ulli( MYSQLAUX_BIND_VARIABLE *var )
{
  return var->buffer.ulli;
}

static double
get_bind_variable_f( MYSQLAUX_BIND_VARIABLE *var )
{
  return var->buffer.f;
}

static double
get_bind_variable_d( MYSQLAUX_BIND_VARIABLE *var )
{
  return var->buffer.d;
}

static char *
get_bind_variable_cv_ptr( MYSQLAUX_BIND_VARIABLE *var )
{
  return var->buffer.cv;
}

static void
get_bind_variable_cv( MYSQLAUX_BIND_VARIABLE *var, char *buf )
{
  memcpy( buf, var->cv, var->length );
}
<#

;;

(define (mysqlaux-type-fixed-size typ)
  (select typ
		[(mysql-type-tiny)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.sc )" unsigned-int)]
		[(mysql-type-short)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.si )" unsigned-int)]
		[(mysql-type-long)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.i )" unsigned-int)]
		[(mysql-type-int24)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.i )" unsigned-int)]
		[(mysql-type-float)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.f )" unsigned-int)]
		[(mysql-type-double)		  (foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.d )" unsigned-int)]
		[(mysql-type-null)				0]
		[(mysql-type-timestamp)		(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.t )" unsigned-int)]
		[(mysql-type-longlong)		(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.lli )" unsigned-int)]
		[(mysql-type-date)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.t )" unsigned-int)]
		[(mysql-type-time)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.t )" unsigned-int)]
		[(mysql-type-datetime)		(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.t )" unsigned-int)]
		[(mysql-type-newdate)			(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.t )" unsigned-int)]
		[(mysql-type-year)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.si )" unsigned-int)]
		[(mysql-type-bit)				  (foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-decimal)		  (foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-newdecimal)  (foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-enum)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-set)				  (foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.??? )" unsigned-int)]
		[(mysql-type-tiny-blob)		(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-medium-blob)	(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-long-blob)		(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-blob)				(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-varchar)			(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-var-string)	(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-string)			(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
		[(mysql-type-geometry)		(foreign-value "sizeof( MYSQLAUX_BIND_VARIABLE_BUFFER.cv )" unsigned-int)]
    [else
      (error ???)] ) )

;; ???

(define (mysql-foreign-type-length foreign-type #!optional (obj (void)))
 (void) )

;; ???

(define (mysql-determine-foreign-type type is-unsigned is-null)
 (void) )

;; ???

(define (mysql-determine-type obj is-unsigned is-null)
 (void) )

;;

(define (mysql-bind-param-init-direct conn idx obj #!key type is-unsigned is-null error)
  (let ([bndptr (mysql-bind-ref-ptr (mysql-connection-binding conn) idx)])
    (mysql-bind-clear! bndptr)
    (unless type
      (let-values ([(typ uflg nflg) (mysql-determine-type obj is-unsigned is-null)])
        (set! type typ)
        (set! is-unsigned uflg)
        (set! is-null nflg) ) )
    (when (and (mysql-null-object? obj) (not (eqv? mysql-type-null type)))
      (warning "null object implies null type")
      (set! type mysql-type-null) )
    (mysql-bind-buffer-type-set! bndptr type)
    (unless (eqv? mysql-type-null type)
      (mysql-bind-is-unsigned-set! bndptr is-unsigned)
      (let* ([foreign-type (mysql-determine-foreign-type type is-unsigned is-null)]
             [len (mysql-foreign-type-length foreign-type obj)])
        (add-bind-variable conn (mysql-bind-init bndptr len is-null)) ) ) ) )

;;

(define (mysql-bind-result-init-direct conn idx type #!optional len)
  (let ([bndptr (mysql-bind-ref-ptr (mysql-connection-binding conn) idx)])
    (mysql-bind-clear! bndptr)
    (mysql-bind-buffer-type-set! bndptr type)
    (unless (eqv? mysql-type-null type)
      (let* ([foreign-type (mysql-determine-foreign-type type #f #f)]
             [len (or len
                      (mysql-foreign-type-length foreign-type))])
        (add-bind-variable conn (mysql-bind-init bndptr len #f)) ) ) ) )

;; Convert the result binding to a Scheme object of typedesc
;FIXME

(define (mysql-bind-result->object bndptr typedesc)
	(void) )

;;

(define (mysql-bindings-reset conn)
  (free-mysql-binding-variables (mysql-connection-bindings conn))
  (mysql-connection-bindings-set! conn #f) )

;; Set the statement attribute

(define (%mysql-stmt-attr-set! stmtptr attr val)
	(cond [(boolean? val)
					(foreign-mysqlaux-stmt-attr-set-bool stmtptr attr val)]
				[(number? val)
					(foreign-mysqlaux-stmt-attr-set-ulong stmtptr attr val)]
				[else
					#t ] ) )

;; Get the statement attribute

(define (%mysql-stmt-attr-get stmtptr attr)
	(select attr
		[(stmt-attr-cursor-type stmt-attr-prefetch-rows)
			(let-location ([val unsigned-long])
				(foreign-mysql-stmt-attr-get stmtptr attr #$val)
				val ) ]
		[(stmt-attr-update-max-length)
			(let-location ([val my-bool])
				(foreign-mysql-stmt-attr-get stmtptr attr #$val)
				val ) ]
		[else
			(void) ] ) )

;; Is the object a NULL object?
;FIXME

(define (mysql-null-object? obj)
	(eq? obj (void)) )

;; Raise a statement expection

(define (signal-mysql-stmt-error loc conn . args)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(let ([err (or (mysql-stmt-error stmtptr)
									 (mysql-stmt-errno stmtptr))]
					[sta (mysql-stmt-sqlstate conn)])
			(apply signal-mysql-condition loc
																		(string-append err
																									 (if sta
																											 (string-append " - " sta)
																											 ""))
																		conn args) ) ) )

;-----------------------------------------------------------------------
; The prepared statement binding MySQL/Scheme API.
;
; This API provides binding construction support.
;

;;
;;	 ((list obj #:type T #:length L #:unsigned B #:null B) ; param 0
;;		 ...
;;		 (list obj #:type T #:length L #:unsigned B #:null B))

(define (make-mysql-param-bindings conn . inits)
)

;;
;;		((list obj #:type T #:length L) ; result 0
;;		 ...
;;		 (list obj #:type T #:length L))

(define (make-mysql-result-bindings conn . inits)
  ; Result metadata available?
  ;
)

;-----------------------------------------------------------------------
; The prepared statement MySQL/Scheme API.
;
; This API provides statement evaluation phase support.
;

;; Frees statement storage

(define (mysql-stmt-close conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(mysql-connection-statement-set! conn #f)
		(when (foreign-mysql-stmt-close stmtptr)
			(signal-mysql-stmt-error 'mysql-stmt-close conn) ) )
  (mysql-bindings-reset conn) )

;; Creates statement storage

(define (mysql-stmt-init conn)
	(mysql-stmt-close conn)
	(let ([stmtptr (foreign-mysql-stmt-init (mysql-connection-ptr conn))])
		(if stmtptr
				(mysql-connection-statement-set! conn stmtptr)
				(signal-mysql-condition 'mysql-stmt-init "out of memory") ) ) )

;;

(define (mysql-stmt-errno conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(foreign-mysql-stmt-errno stmtptr) ) )

;; Returns a string describing the last mysql stmt error, or #f if no error
;; has occurred.

(define (mysql-stmt-error conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(let ([errstr (foreign-mysql-stmt-error stmtptr)])
			(and (not (string=? "" errstr))
					 errstr ) ) ) )

;; Returns a string describing the last mysql stmt state error, or #f if no error
;; has occurred.

(define (mysql-stmt-sqlstate conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(let ([errstr (foreign-mysql-stmt-sqlstate stmtptr)])
			(and (not (or (string=? "00000" errstr)
										(string=? "HY000" errstr)))
					 errstr ) ) ) )

;;

(define (mysql-stmt-prepare conn sql)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(unless (fx= 0 (foreign-mysql-stmt-prepare stmtptr sql (string-length sql)))
			(signal-mysql-stmt-error 'mysql-stmt-prepare conn) ) ) )

;;

(define (mysql-stmt-param-count conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(foreign-mysql-stmt-param-count stmtptr) ) )

;; Bind the fully defined parameter binding

(define (mysql-stmt-bind-param-direct conn binding)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(when (foreign-mysql-stmt-bind-param stmtptr (mysql-binding-binds binding))
      (signal-mysql-stmt-error 'mysql-stmt-bind-param-direct conn) ) ) )

;; Bind the fully defined result binding

(define (mysql-stmt-bind-result-direct conn binding)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(when (foreign-mysql-stmt-bind-result stmtptr (mysql-binding-binds binding))
      (signal-mysql-stmt-error 'mysql-stmt-bind-result-direct conn) ) ) )

;; Toggles the connection binding to params

(define (mysql-stmt-bind-param conn binding)
  (mysql-bindings-reset conn)
  (mysql-stmt-bind-param-direct conn binding)
	(mysql-connection-bindings-set! conn binding) )

;; Toggles the connection binding to results

(define (mysql-stmt-bind-result conn binding)
  (mysql-bindings-reset conn)
  (mysql-stmt-bind-result-direct conn binding)
	(mysql-connection-bindings-set! conn binding) )

;;

(define (mysql-stmt-execute conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(unless (fx= 0 (foreign-mysql-stmt-execute stmtptr))
			(signal-mysql-stmt-error 'mysql-stmt-execute conn) ) ) )

;;

(define (mysql-stmt-affected-rows conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(let ([cnt (foreign-mysql-stmt-affected-rows stmtptr)])
			(and (not (= -1 cnt))
					 cnt ) ) ) )

;; Returns boolean for success, mysql-data-truncated, or signals
;; an exception.

(define (mysql-stmt-fetch conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(let ([val (foreign-mysql-stmt-fetch stmtptr)])
			(cond [(fx= 0 val)
							#t]
						[(= mysql-no-data val)
							#f]
						[(= mysql-data-truncated val)
							mysql-data-truncated]
						[(= 1 val)
							(signal-mysql-stmt-error 'mysql-stmt-fetch conn) ] ) ) ) )

;; Causes the result to be buffered. Does not touch the connection
;; result!

(define (mysql-stmt-store-result conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(when (fx= 0 (foreign-mysql-stmt-store-result stmtptr))
			(signal-mysql-stmt-error 'mysql-stmt-store-result conn) ) ) )

;; Can only be invoked after a stmt-store-result, stmt-fetch, or
;; stmt-prepare (of a result).

(define (mysql-stmt-result-metadata conn)
	(mysql-free-result conn) ; free any existing results
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(let ([resptr (foreign-mysql-stmt-result-metadata stmtptr)])
			(cond [resptr
              (mysql-connection-result-set! conn resptr)
              (mysql-connection-result-start-set!
               conn
               (foreign-mysql-stmt-row-tell (mysql-connection-result conn))) ]
            [else
              (signal-mysql-stmt-error 'mysql-stmt-result-metadata conn) ] ) ) ) )

;; Set the connection statement attribute

(define (mysql-stmt-attr-set! conn attr val)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(when (%mysql-stmt-attr-set! stmtptr attr val)
			(signal-mysql-condition 'mysql-stmt-attr-set
															"unknown statement attribute" attr val) ) ) )

;; Get the connection statement attribute

(define (mysql-stmt-attr-get conn attr)
	(and-let* ([stmtptr (mysql-connection-statement conn)])
		(let ([val (%mysql-stmt-attr-get stmtptr attr)])
			(if (eq? (void) val)
					(signal-mysql-condition 'mysql-stmt-attr-get
																	 "unknown statement attribute" attr)
					val ) ) ) )

;; Rewinds to the beginning of the result set. Has no effect if there is no
;; current result set.

(define (mysql-stmt-rewind conn)
	(and-let* ([stmtptr (mysql-connection-statement conn)]
						 [resptr (mysql-connection-result-start conn)])
		(foreign-mysql-stmt-row-seek stmtptr resptr) ) )

;-----------------------------------------------------------------------
; The prepared statement "extended" MySQL/Scheme API.
;
; This API provides statement query support.
;

;; Returns a procedure, or #f when no connection.
;; The procedure takes a field identifier and returns the
;; mysql-bind-ptr of the field's mysql-bind record, or #f
;; when no more rows to fetch.

(define (mysql-stmt-row-fetch conn)
	(and-let* ([resptr (mysql-connection-result conn)]      ; metadata, if any
	           [bindings (mysql-connection-bindings conn)]  ; better be result
						 [bndptr (mysql-binding-binds bindings)]
						 [(mysql-stmt-fetch conn)])
		(let ([fldcnt (mysql-num-fields conn)])
			(lambda (field)
				(and-let* ([fldidx (mysql-get-field-index resptr field fldcnt)])
					(mysql-bind-ref-ptr bndptr fldidx) ) ) ) ) )

;; Fetch statement result metadata for query?
;; Necessary for symbolic field name resolution

(define-parameter mysql-stmt-fetch-metadata-for-query
  #t
  (lambda (x) x) )

;; Prep a query
;; query    : SQL source string

(define (mysql-stmt-query-init conn query)
  ; Must have statement storage
  (unless (mysql-connection-statement conn) (mysql-stmt-init conn) )
  ; Compile SQL
	(mysql-stmt-prepare conn query)
	; Using metadata? (sets result-set)
	(if (mysql-stmt-fetch-metadata-for-query)
	    (mysql-stmt-result-metadata conn)
	    (mysql-free-result conn) ) )

;; Perform a query
;; query    : SQL source string
;; params   : mysql-binding
;; results  : mysql-binding

(define (mysql-stmt-query conn query #!optional params results)
  ; Prep the stmt
  (mysql-stmt-query-init conn query)
	; Bind any params
	(when params (mysql-stmt-bind-param conn params))
	; Perform
	(mysql-stmt-execute conn)
  ; Bind any results
	(when results (mysql-stmt-bind-result conn results)) )

;-----------------------------------------------------------------------
; The prepared statement "map" MySQL/Scheme API.
;
; This API provides some additional functionality for traversing results
; in a Scheme-ish way.
;

;; Calls proc on every row in the current result set. proc should take 3
;; arguments: the row (as described for mysql-stmt-row-fetch), the row index
;; (which starts with 1 and ends with (mysql-stmt-num-rows conn)), and the
;; current accumulated value.
;;
;; Returns the final accumulated value.
;;
;; Note: rewinds the result before and after iterating over it; thus,
;; all rows are included.
;;
;; You must call mysql-stmt-rewind if you later want to iterate over the result set
;; using mysql-stmt-row-fetch.

(define (mysql-stmt-row-fold conn proc init)
	(mysql-stmt-rewind conn)
	(let loop ([rownum 1] [acc init])
		(let ([row (mysql-stmt-row-fetch conn)])
			(if row
					(loop (fx+ rownum 1) (proc row rownum acc))
					acc ) ) ) )

;; Calls proc on every row in the current result set. proc should take 2
;; arguments: the row (as described for mysql-stmt-row-fetch) and the row index
;; (which starts with 1 and ends with (mysql-stmt-num-rows conn)).
;;
;; Note: rewinds the result before and after iterating over it; thus,
;; all rows are included.
;;
;; You must call mysql-stmt-rewind if you later want to iterate over the result set
;; using mysql-stmt-row-fetch.

(define (mysql-stmt-row-for-each conn proc)
	(mysql-stmt-row-fold conn
											 (lambda (row rownum _) (proc row rownum))
											 #t) )

;; Calls proc on every row in the current result set. proc should take 2
;; arguments: the row (as described for mysql-stmt-row-fetch) and the row index
;; (which starts with 1 and ends with (mysql-stmt-num-rows conn)).
;;
;; Returns a list of the results of each proc invocation.
;;
;; Note: rewinds the result before and after iterating over it; thus,
;; all rows are included.
;;
;; You must call mysql-stmt-rewind if you later want to iterate over the result set
;; using mysql-stmt-row-fetch.

(define (mysql-stmt-row-map conn proc)
	(reverse!
	 (mysql-stmt-row-fold conn
												(lambda (row rownum lst) (cons (proc row rownum) lst))
												'())) )

;; Executes query and then mysql-row-for-each with the given proc. The proc
;; must meet the contract specified for the proc passed to mysql-stmt-row-fold.

(define (mysql-stmt-query-fold conn query proc init #!optional params results)
	(mysql-stmt-query conn query params results)
	(mysql-stmt-row-fold conn proc init) )

;; Executes query and then mysql-row-for-each with the given proc. The proc
;; must meet the contract specified for the proc passed to mysql-stmt-row-for-each.

(define (mysql-stmt-query-for-each conn query proc #!optional params results)
	(mysql-stmt-query conn query params results)
	(mysql-stmt-row-for-each conn proc) )

;; Executes query and then mysql-row-for-each with the given proc. The proc
;; must meet the contract specified for the proc passed to mysql-stmt-row-map.

(define (mysql-stmt-query-map conn query proc #!optional params results)
	(mysql-stmt-query conn query params results)
	(mysql-stmt-row-map conn proc) )
|#
